"""Validation of configuration."""

import logging
import os
from pathlib import Path

from schema import And, Optional, Or, Schema, SchemaError

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=line-too-long
# pylint: disable=too-many-lines


def validate_config(config: dict, schema: Schema, config_type: str) -> None:
    """
    Validate configuration.

    Parameters
    ----------
    config : dict
        Config dictionary imported by read_yaml() and parsed through clean_config().
    schema : Schema
        A schema against which the configuration is to be compared.
    config_type : str
        Description of of configuration being validated.
    """
    try:
        schema.validate(config)
        LOGGER.info(f"The {config_type} is valid.")
    except SchemaError as schema_error:
        raise SchemaError(
            f"There is an error in your {config_type} configuration. "
            "Please refer to the first error message above for details"
        ) from schema_error


DEFAULT_CONFIG_SCHEMA = Schema(
    {
        "base_dir": Path,
        "output_dir": Path,
        "log_level": Or(
            "debug",
            "info",
            "warning",
            "error",
            error="Invalid value in config for 'log_level', valid values are 'info' (default), 'debug', 'error' or 'warning",
        ),
        "cores": lambda n: 1 <= n <= os.cpu_count(),
        "file_ext": Or(
            ".spm",
            ".asd",
            ".jpk",
            ".ibw",
            ".gwy",
            ".topostats",
            ".stp",
            ".top",
            error="Invalid value in config for 'file_ext', valid values are '.spm', '.jpk', '.ibw', '.gwy', '.topostats', or '.asd'.",
        ),
        "output_stats": Or(
            "basic", "full", error="Invalid value for 'output_stats', valid values are 'full' or 'basic'."
        ),
        "loading": {
            "channel": str,
            "extract": Or(
                "all",
                "raw",
                "filters",
                error="Invalid value in config for 'extract', valid values are 'all', 'raw' or 'filters'",
            ),
        },
        "filter": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'filter.run', valid values are 'True' or 'False'",
            ),
            "row_alignment_quantile": lambda n: 0.0 <= n <= 1.0,
            "threshold_method": Or(
                "absolute",
                "otsu",
                "std_dev",
                error=(
                    "Invalid value in config for 'filter.threshold_method', valid values "
                    "are 'absolute', 'otsu' or 'std_dev'"
                ),
            ),
            "otsu_threshold_multiplier": float,
            "threshold_std_dev": {
                "below": lambda n: n > 0,
                "above": lambda n: n > 0,
            },
            "threshold_absolute": {
                "below": Or(
                    int,
                    float,
                    error=("Invalid value in config for filter.threshold.absolute.below should be type int or float"),
                ),
                "above": Or(
                    int,
                    float,
                    error=("Invalid value in config for filter.threshold.absolute.below should be type int or float"),
                ),
            },
            "gaussian_size": float,
            "gaussian_mode": Or(
                "nearest",
                error="Invalid value in config for 'filter.gaussian_mode', valid values are 'nearest'",
            ),
            "remove_scars": {
                "run": bool,
                "removal_iterations": lambda n: 0 <= n < 10,
                "threshold_low": lambda n: n > 0,
                "threshold_high": lambda n: n > 0,
                "max_scar_width": lambda n: n >= 1,
                "min_scar_length": lambda n: n >= 1,
            },
        },
        "grains": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for grains.run, valid values are 'True' or 'False'",
            ),
            "grain_crop_padding": int,
            "threshold_method": Or(
                "absolute",
                "otsu",
                "std_dev",
                error=(
                    "Invalid value in config for 'grains.threshold_method', valid values "
                    "are 'absolute', 'otsu' or 'std_dev'"
                ),
            ),
            "otsu_threshold_multiplier": float,
            "threshold_std_dev": {
                "below": [lambda n: n > 0],
                "above": [
                    lambda n: n > 0,
                ],
            },
            "threshold_absolute": {
                "below": [
                    Or(
                        int,
                        float,
                        error=(
                            "Invalid value in config for grains.threshold.absolute.below should be type int or float"
                        ),
                    )
                ],
                "above": [
                    Or(
                        int,
                        float,
                        error=(
                            "Invalid value in config for grains.threshold.absolute.above should be type int or float"
                        ),
                    )
                ],
            },
            "area_thresholds": {
                "above": [
                    Or(
                        int,
                        None,
                        error=(
                            "Invalid value in config for 'grains.area_thresholds.above', valid values are int or null"
                        ),
                    )
                ],
                "below": [
                    Or(
                        int,
                        None,
                        error=(
                            "Invalid value in config for 'grains.area_thresholds.below', valid values are int or null"
                        ),
                    )
                ],
            },
            "direction": Or(
                "both",
                "below",
                "above",
                error="Invalid direction for grains.direction valid values are 'both', 'below' or 'above",
            ),
            "remove_edge_intersecting_grains": Or(
                True,
                False,
                error="Invalid value in config for 'grains.remove_edge_intersecting_grains', valid values are 'True' or 'False'",
            ),
            "unet_config": {
                "model_path": Or(None, str),
                "upper_norm_bound": float,
                "lower_norm_bound": float,
                "remove_disconnected_grains": bool,
                "confidence": float,
            },
            "vetting": {
                "whole_grain_size_thresholds": Or(
                    None,
                    [lambda n: n > 0, lambda n: n > 0],
                    error="Invalid value in config for 'grainstats.vetting.whole_grain_size_thresholds', this needs to"
                    "be a list of two positive floats",
                ),
                "class_conversion_size_thresholds": Or(
                    None,
                    # List of lists of 3 integers and 2 integers
                    [[[int, int, int], [int, int]]],
                    error="Invalid value in config for 'grainstats.vetting.class_conversion_size_thresholds', this "
                    "needs to be a list of lists of 3 integers and 2 integers - [class number, class to convert to "
                    "if too small, class to convert to if too big] and [minimum size, maximum size]",
                ),
                "class_region_number_thresholds": Or(
                    None,
                    # List of lists of 3 integers
                    [[int, int, int]],
                    error="Invalid value in config for 'grainstats.vetting.class_region_number_thresholds', this needs"
                    "to be a list of lists of 3 integers - class number, minimum region number, maximum region number",
                ),
                "class_size_thresholds": Or(
                    None,
                    # List of lists of 3 integers
                    [[int, int, int]],
                    error="Invalid value in config for 'grainstats.vetting.class_size_thresholds', this needs to be a"
                    "list of lists of 3 integers - class number, minimum size, maximum size",
                ),
                "nearby_conversion_classes_to_convert": Or(
                    None,
                    # List of tuples of two integers
                    [[int, int]],
                    error="Invalid value in config for 'grainstats.vetting.nearby_conversion_classes_to_convert', this"
                    "needs to be a list of tuples of two integers",
                ),
                "class_touching_threshold": lambda n: n >= 0,
                "keep_largest_labelled_regions_classes": Or(
                    None,
                    # List of integers
                    [int],
                    error="Invalid value in config for 'grainstats.vetting.keep_largest_labelled_regions_classes', this"
                    "needs to be a list of integers",
                ),
                "class_connection_point_thresholds": Or(
                    None,
                    # List of tuples: list[tuple[[tuple[int, int], tuple[int, int]]]
                    [[[int, int], [int, int]]],
                    error="Invalid value in config for 'grainstats.vetting.class_connection_point_thresholds', this"
                    "needs to be a list of tuples of two tuples of two integers. Eg [((1, 2), (3, 4))]",
                ),
            },
            "classes_to_merge": Or(
                None,
                [int],
                error="Invalid value in config for 'grains.classes_to_merge', this needs to be a list of integers.",
            ),
        },
        "grainstats": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'grainstats.run', valid values are 'True' or 'False'",
            ),
            "edge_detection_method": Or(
                "binary_erosion",
                "canny",
            ),
            "extract_height_profile": Or(
                True,
                False,
                error="Invalid value in config for 'grainstats.extract_height_profile',"
                "valid values are 'True' or 'False'",
            ),
            "class_names": [str],
        },
        "disordered_tracing": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'disordered_tracing.run', valid values are 'True' or 'False'",
            ),
            "class_index": int,
            "min_skeleton_size": lambda n: n > 0.0,
            "mask_smoothing_params": {
                "gaussian_sigma": Or(
                    float,
                    int,
                    None,
                ),
                "dilation_iterations": Or(
                    int,
                    None,
                ),
                "holearea_min_max": [
                    Or(
                        int,
                        float,
                        None,
                        error=(
                            "Invalid value in config for 'disordered_tracing.mask_smoothing_params.holearea_min_max', valid values "
                            "are int, float or null"
                        ),
                    ),
                ],
            },
            "skeletonisation_params": {
                "method": Or(
                    "zhang",
                    "lee",
                    "thin",
                    "medial_axis",
                    "topostats",
                    error="Invalid value in config for 'disordered_tracing.skeletonisation_method',"
                    "valid values are 'zhang', 'lee', 'thin', 'medial_axis', 'topostats'",
                ),
                "height_bias": lambda n: 0 < n <= 1,
            },
            "pruning_params": {
                "method": Or(
                    "topostats",
                    error="Invalid value in config for 'disordered_tracing.pruning_method', valid values are 'topostats'",
                ),
                "max_length": lambda n: n >= 0,
                "method_values": Or("min", "median", "mid"),
                "method_outlier": Or("abs", "mean_abs", "iqr"),
                "height_threshold": Or(int, float, None),
                "only_height_prune_endpoints": Or(
                    True,
                    False,
                    error="Invalid value in config for 'disordered_tracing.pruning_params.run', valid values are 'True' or 'False'",
                ),
            },
        },
        "nodestats": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'nodestats.run', valid values are 'True' or 'False'",
            ),
            "node_joining_length": float,
            "node_extend_dist": float,
            "branch_pairing_length": float,
            "pair_odd_branches": bool,
        },
        "ordered_tracing": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'ordered_tracing.run', valid values are 'True' or 'False'",
            ),
            "ordering_method": Or(
                "nodestats",
                "original",
                error="Invalid value in config for 'ordered_tracing.ordering_method', valid values are 'nodestats' or 'original'",
            ),
        },
        "splining": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'splining.run', valid values are 'True' or 'False'",
            ),
            "method": Or(
                "spline",
                "rolling_window",
                error="Invalid value in config for 'splining.method', valid values are 'spline' or 'rolling_window'",
            ),
            "rolling_window_size": lambda n: n > 0.0,
            "rolling_window_resampling": bool,
            "rolling_window_resample_regular_spatial_interval": lambda n: n > 0.0,
            "spline_step_size": lambda n: n > 0.0,
            "spline_linear_smoothing": lambda n: n >= 0.0,
            "spline_circular_smoothing": lambda n: n >= 0.0,
            "spline_degree": int,
            # "cores": lambda n: n > 0.0,
        },
        "curvature": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'curvature.run', valid values are 'True' or 'False'",
            ),
            "colourmap_normalisation_bounds": [
                Or(
                    float,
                    int,
                    error="Invalid value in config for 'curvature.colourmap_normalisation_bounds', valid values"
                    "are float or int",
                )
            ],
        },
        "plotting": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for 'plotting.run', valid values are 'True' or 'False'",
            ),
            "style": And(
                str,
                Or(
                    "topostats.mplstyle",
                    str,
                    Path,
                    None,
                    error="Invalid value in config for 'plotting.style', valid values are 'topostats.mplstyle' or None",
                ),
            ),
            "savefig_format": Or(
                None,
                str,
                error="Invalid value in config for plotting.savefig_formatmust be a value supported by Matplotlib.",
            ),
            "savefig_dpi": Or(
                None,
                "figure",
                lambda n: n > 0,
                error="Invalid value in config for plotting.savefig_dpi, validvalues are 'figure' or floats",
            ),
            "image_set": [
                Or(
                    "all",
                    "core",
                    "filters",
                    "grains",
                    "grain_crops",
                    "disordered_tracing",
                    "nodestats",
                    "ordered_tracing",
                    "splining",
                    error="Invalid value in config for 'plotting.image_set', valid values are 'all' or 'core'",
                )
            ],
            "pixel_interpolation": Or(
                None,
                "none",
                "bessel",
                "bicubic",
                "bilinear",
                "catrom",
                "gaussian",
                "hamming",
                "hanning",
                "hermite",
                "kaiser",
                "lanczos",
                "mitchell",
                "nearest",
                "quadric",
                "sinc",
                "spline16",
                "spline36",
                error="Invalid interpolation value. See https://matplotlib.org/stable/gallery/images_contours_and_fields/interpolation_methods.html for options.",
            ),
            "grain_crop_plot_size_nm": Or(
                int,
                float,
                error=("Invalid value in config for 'grains.grain_crop_plot_size_nm', valid values are int or float"),
            ),
            "zrange": [float, int, None],
            "colorbar": Or(
                True,
                False,
                error="Invalid value in config for 'plotting.colorbar', valid values are 'True' or 'False'",
            ),
            "axes": Or(
                True,
                False,
                error="Invalid value in config plotting.for 'axes', valid values are 'True' or 'False'",
            ),
            "num_ticks": Or(
                [None, And(int, lambda n: n > 1)],
                error="Invalid value in config plotting.for 'num_ticks', valid values are 'null' or integers > 1",
            ),
            "cmap": Or(
                None,
                str,
                error="Invalid value in config for 'plotting.cmap', valid values are 'afmhot', 'nanoscope', "
                "'gwyddion' or values supported by Matplotlib",
            ),
            "mask_cmap": str,
            "histogram_log_axis": Or(
                True,
                False,
                error=(
                    "Invalid value in config plotting histogram. For 'log_y_axis', valid values are 'True' or 'False'"
                ),
            ),
            "number_grains": Or(
                True,
                False,
                error=("Invalid value in config for 'plotting.number_grain_plots', valid values are 'True' or 'False'"),
            ),
        },
        "summary_stats": {
            "run": Or(
                True,
                False,
                error="Invalid value in config for summary_stats.run, valid values are 'True' or 'False'",
            ),
            "config": Or(
                None,
                str,
                error=(
                    "Invalid value in config for summary_stats.config, valid values are 'None' or a path to a "
                    "config file."
                ),
            ),
        },
    }
)


PLOTTING_SCHEMA = Schema(
    {
        "extracted_channel": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'extracted_channel.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": Or(
                "filters",
                "grains",
                "grain_crops",
                "disordered_tracing",
                "nodestats",
                "ordered_tracing",
                "splining",
                error=(
                    "Invalid value in config 'extracted_channel.module', valid values "
                    "are 'filters', 'grains', 'grain_crops', 'disordered_tracing', "
                    "'nodestats', 'ordered_tracing', or 'splining'"
                ),
            ),
        },
        "pixels": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error="Invalid value in config 'pixels.image_type', valid values are 'binary' or 'non-binary'",
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": Or(
                "filters",
                "grains",
                "grain_crops",
                "disordered_tracing",
                "nodestats",
                "ordered_tracing",
                "splining",
                error=(
                    "Invalid value in config 'pixels.module', valid values "
                    "are 'filters', 'grains', 'grain_crops', 'disordered_tracing', "
                    "'nodestats', 'ordered_tracing', or 'splining'"
                ),
            ),
        },
        "initial_median_flatten": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_median_flatten.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": Or(
                "filters",
                "grains",
                "grain_crops",
                "disordered_tracing",
                "nodestats",
                "ordered_tracing",
                "splining",
                error=(
                    "Invalid value in config 'initial_median_flatten.module', valid values "
                    "are 'filters', 'grains', 'grain_crops', 'disordered_tracing', "
                    "'nodestats', 'ordered_tracing', or 'splining'"
                ),
            ),
        },
        "initial_tilt_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_tilt_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": Or(
                "filters",
                "grains",
                "grain_crops",
                "disordered_tracing",
                "nodestats",
                "ordered_tracing",
                "splining",
                error=(
                    "Invalid value in config 'initial_tilt_removal.module', valid values "
                    "are 'filters', 'grains', 'grain_crops', 'disordered_tracing', "
                    "'nodestats', 'ordered_tracing', or 'splining'"
                ),
            ),
        },
        "initial_quadratic_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_quadratic_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "initial_scar_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_scar_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "initial_zero_average_background": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_zero_average_background.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "initial_nonlinear_polynomial_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'initial_nonlinear_polynomial_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "mask": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error="Invalid value in config 'mask.image_type', valid values are 'binary' or 'non-binary'",
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "masked_median_flatten": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_median_flatten.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "masked_tilt_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_tilt_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "masked_quadratic_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_quadratic_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "masked_nonlinear_polynomial_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_nonlinear_polynomial_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "secondary_scar_removal": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_quadratic_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "scar_mask": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'masked_quadratic_removal.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "final_zero_average_background": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'final_zero_average_background.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "gaussian_filtered": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'gaussian_filtered.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "z_threshed": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'z_threshold.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": True,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "filters",
        },
        "thresholded_grains": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'mask_grains.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "mask_cmap": str,
            "module": "grains",
        },
        "tidied_border": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'tidied_border.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "mask_cmap": str,
            "module": "grains",
        },
        "removed_objects_too_small_to_process": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'removed_small_objects.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "area_thresholded": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'removed_objects_too_small_to_process.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "unet": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'unet_tensor.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "vetted": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'vetted_tensor.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "merged_classes": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'merged_classes_tensor.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "mask_overlay": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'mask_overlay.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": True,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "bounding_boxes": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'bounding_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grains",
        },
        "grain_image": {
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'grain_image.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": False,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grain_crops",
        },
        "grain_mask": {
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'grain_mask.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grain_crops",
        },
        "grain_mask_image": {
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'grain_mask_image.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "grain_crops",
        },
        "orig_grain": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "module": "disordered_tracing",
        },
        "smoothed_grain": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "module": "disordered_tracing",
        },
        "skeleton": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "disordered_tracing",
        },
        "pruned_skeleton": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "disordered_tracing",
        },
        "branch_indexes": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'branch_indexes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "disordered_tracing",
        },
        "branch_types": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "disordered_tracing",
        },
        "convolved_skeletons": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'convolved_skeleton.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "nodestats",
        },
        "node_centres": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'node_centres.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "nodestats",
        },
        "connected_nodes": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'node_branch_mask.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "nodestats",
        },
        "node_area_skeleton": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'node_area_skeleton.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "nodestats",
        },
        "node_branch_mask": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'node_branch_mask.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "nodestats",
        },
        "node_avg_mask": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'node_avg_mask.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "module": "nodestats",
        },
        "node_line_trace": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "module": "nodestats",
        },
        "ordered_traces": {
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'all_molecule_traces.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "ordered_tracing",
        },
        "trace_segments": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'all_molecule_traces.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "ordered_tracing",
        },
        "over_under": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'all_molecule_traces.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "ordered_tracing",
        },
        "all_molecules": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "ordered_tracing",
        },
        "fitted_trace": {
            "filename": str,
            "title": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'coloured_boxes.image_type', valid values are 'binary' or 'non-binary'"
                ),
            ),
            "mask_cmap": str,
            "core_set": bool,
            "savefig_dpi": int,
            "module": "ordered_tracing",
        },
        "splined_trace": {
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'splined_trace.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "title": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "splining",
        },
        "curvature": {
            "image_type": Or(
                "binary",
                "non-binary",
                error=("Invalid value in config 'curvature.image_type', valid values are 'binary' or 'non-binary'"),
            ),
            "title": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "splining",
        },
        "curvature_individual_grains": {
            "filename": str,
            "image_type": Or(
                "binary",
                "non-binary",
                error=(
                    "Invalid value in config 'curvature_individual_grains.image_type', valid values "
                    "are 'binary' or 'non-binary'"
                ),
            ),
            "title": str,
            "core_set": bool,
            "savefig_dpi": Or(
                lambda n: n > 0,
                "figure",
                error="Invalid value in config for 'dpi', valid values are 'figure' or > 0.",
            ),
            "module": "splining",
        },
    }
)

SUMMARY_SCHEMA = Schema(
    {
        "base_dir": Path,
        "output_dir": Path,
        "csv_file": str,
        "savefig_format": Or(
            "png",
            "pdf",
            "svg",
            "tiff",
            "tif",
            error=("Invalid value in config 'savefig_format', valid values are 'png', 'pdf', 'svg', 'tiff' or 'tif'"),
        ),
        "var_to_label": Or(
            None,
            str,
            error="Invalid value in config for 'var_to_label', valid values are 'None' or a str",
        ),
        "image_id": str,
        "molecule_id": str,
        "hist": Or(
            True,
            False,
            error="Invalid value in config for 'hist', valid values are 'True' or 'False'",
        ),
        "bins": lambda n: n > 0,
        "stat": Or(
            "count",
            "frequency",
            "probability",
            "percent",
            "density",
            error=(
                "Invalid value in config 'stat', valid values are 'count', 'frequency', "
                "'probability', 'percent' or 'density'"
            ),
        ),
        "kde": Or(
            True,
            False,
            error="Invalid value in config for 'kde', valid values are 'True' or 'False'",
        ),
        "violin": Or(
            True,
            False,
            error="Invalid value in config for 'violin', valid values are 'True' or 'False'",
        ),
        "figsize": [lambda n: n > 0],
        "alpha": lambda n: n > 0,
        "palette": Or(
            "colorblind",
            "deep",
            "muted",
            "pastel",
            "bright",
            "dark",
            "Spectral",
            "Set2",
            error=(
                "Invalid value in config 'palette', valid values are 'colorblind', 'deep', "
                "'muted', 'pastel', 'bright', 'dark', 'Spectral' or 'Set2'"
            ),
        ),
        "stats_to_sum": [
            Optional("area"),
            Optional("area_cartesian_bbox"),
            Optional("aspect_ratio"),
            Optional("bending_angle"),
            Optional("total_contour_length"),
            Optional("average_end_to_end_distance"),
            Optional("height_max"),
            Optional("height_mean"),
            Optional("height_median"),
            Optional("height_min"),
            Optional("max_feret"),
            Optional("min_feret"),
            Optional("radius_max"),
            Optional("radius_mean"),
            Optional("radius_median"),
            Optional("radius_min"),
            Optional("smallest_bounding_area"),
            Optional("smallest_bounding_length"),
            Optional("smallest_bounding_width"),
            Optional("volume"),
        ],
    }
)



"""Order single pixel skeletons with or without NodeStats Statistics."""

import logging
import math

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import interpolate as interp

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)

# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
# pylint: disable=too-many-positional-arguments


# pylint: disable=too-many-instance-attributes
class splineTrace:
    # pylint: disable=too-many-instance-attributes
    """
    Smooth the ordered trace via an average of splines.

    Parameters
    ----------
    image : npt.NDArray
        Whole image containing all molecules and grains.
    mol_ordered_tracing_data : dict
        Molecule ordered trace dictionary containing Nx2 ordered coords and molecule statistics.
    pixel_to_nm_scaling : float
        The pixel to nm scaling factor, by default 1.
    spline_step_size : float
        Step length in meters to use a coordinate for splining.
    spline_linear_smoothing : float
        Amount of linear spline smoothing.
    spline_circular_smoothing : float
        Amount of circular spline smoothing.
    spline_degree : int
        Degree of the spline. Cubic splines are recommended. Even values of k should be avoided especially with a
        small s-value.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        image: npt.NDArray,
        mol_ordered_tracing_data: dict,
        pixel_to_nm_scaling: float,
        spline_step_size: float,
        spline_linear_smoothing: float,
        spline_circular_smoothing: float,
        spline_degree: int,
    ) -> None:
        # pylint: disable=too-many-arguments
        """
        Initialise the splineTrace class.

        Parameters
        ----------
        image : npt.NDArray
            Whole image containing all molecules and grains.
        mol_ordered_tracing_data : dict
            Nx2 ordered trace coordinates.
        pixel_to_nm_scaling : float
            The pixel to nm scaling factor, by default 1.
        spline_step_size : float
            Step length in meters to use a coordinate for splining.
        spline_linear_smoothing : float
            Amount of linear spline smoothing.
        spline_circular_smoothing : float
            Amount of circular spline smoothing.
        spline_degree : int
            Degree of the spline. Cubic splines are recommended. Even values of k should be avoided especially with a
            small s-value.
        """
        self.image = image
        self.number_of_rows, self.number_of_columns = image.shape
        self.mol_ordered_trace = mol_ordered_tracing_data["ordered_coords"]
        self.mol_is_circular = mol_ordered_tracing_data["mol_stats"]["circular"]
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.spline_step_size = spline_step_size
        self.spline_linear_smoothing = spline_linear_smoothing
        self.spline_circular_smoothing = spline_circular_smoothing
        self.spline_degree = spline_degree

        self.tracing_stats = {
            "contour_length": None,
            "end_to_end_distance": None,
        }

    def get_splined_traces(
        self,
        fitted_trace: npt.NDArray,
    ) -> npt.NDArray:
        """
        Get a splined version of the fitted trace - useful for finding the radius of gyration etc.

        This function actually calculates the average of several splines which is important for getting a good fit on
        the lower resolution data.

        Parameters
        ----------
        fitted_trace : npt.NDArray
            Numpy array of the fitted trace.

        Returns
        -------
        npt.NDArray
            Splined (smoothed) array of trace.
        """
        # Calculate the step size in pixels from the step size in metres.
        # Should always be at least 1.
        # Note that step_size_m is in m and pixel_to_nm_scaling is in m because of the legacy code which seems to almost
        # always have pixel_to_nm_scaling be set in metres using the flag convert_nm_to_m. No idea why this is the case.
        step_size_px = max(int(self.spline_step_size / (self.pixel_to_nm_scaling * 1e-9)), 1)
        # Splines will be totalled and then divived by number of splines to calculate the average spline
        spline_sum = None

        # Get the length of the fitted trace
        fitted_trace_length = fitted_trace.shape[0]

        # If the fitted trace is less than the degree plus one, then there is no
        # point in trying to spline it, just return the fitted trace
        if fitted_trace_length < self.spline_degree + 1:
            LOGGER.debug(
                f"Fitted trace for grain {step_size_px} too small ({fitted_trace_length}), returning fitted trace"
            )

            return fitted_trace

        # There cannot be fewer than degree + 1 points in the spline
        # Decrease the step size to ensure more than this number of points
        while fitted_trace_length / step_size_px < self.spline_degree + 1:
            # Step size cannot be less than 1
            if step_size_px <= 1:
                step_size_px = 1
                break
            step_size_px = -1

        # Set smoothness and periodicity appropriately for linear / circular molecules.
        spline_smoothness, spline_periodicity = (
            (self.spline_circular_smoothing, 2) if self.mol_is_circular else (self.spline_linear_smoothing, 0)
        )

        # Create an array of evenly spaced points between 0 and 1 for the splines to be evaluated at.
        # This is needed to ensure that the splines are all the same length as the number of points
        # in the spline is controlled by the ev_array variable.
        ev_array = np.linspace(0, 1, fitted_trace_length * step_size_px)

        # Find as many splines as there are steps in step size, this allows for a better spline to be obtained
        # by averaging the splines. Think of this like weaving a lot of splines together along the course of
        # the trace. Example spline coordinate indexes: [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4], where spline
        # 1 takes every 4th coordinate, starting at position 0, then spline 2 takes every 4th coordinate
        # starting at position 1, etc...
        for i in range(step_size_px):
            # Sample the fitted trace at every step_size_px pixels
            sampled = [fitted_trace[j, :] for j in range(i, fitted_trace_length, step_size_px)]

            # Scipy.splprep cannot handle duplicate consecutive x, y tuples, so remove them.
            # Get rid of any consecutive duplicates in the sampled coordinates
            sampled = self.remove_duplicate_consecutive_tuples(tuple_list=sampled)

            x_sampled = sampled[:, 0]
            y_sampled = sampled[:, 1]

            # Use scipy's B-spline functions
            # tck is a tuple, (t,c,k) containing the vector of knots, the B-spline coefficients
            # and the degree of the spline.
            # s is the smoothing factor, per is the periodicity, k is the degree of the spline
            tck = interp.splprep(
                [x_sampled, y_sampled],
                s=spline_smoothness,
                per=spline_periodicity,
                # quiet=self.spline_quiet,
                k=self.spline_degree,
            )[0]
            # splev returns a tuple (x_coords ,y_coords) containing the smoothed coordinates of the
            # spline, constructed from the B-spline coefficients and knots. The number of points in
            # the spline is controlled by the ev_array variable.
            # ev_array is an array of evenly spaced points between 0 and 1.
            # This is to ensure that the splines are all the same length.
            # Tck simply provides the coefficients for the spline.
            out = interp.splev(ev_array, tck)
            splined_trace = np.column_stack((out[0], out[1]))

            # Add the splined trace to the spline_sum array for averaging later
            if spline_sum is None:
                spline_sum = np.array(splined_trace)
            else:
                spline_sum = np.add(spline_sum, splined_trace)

        # Find the average spline between the set of splines
        # This is an attempt to find a better spline by averaging our candidates

        return np.divide(spline_sum, [step_size_px, step_size_px])

    @staticmethod
    # Perhaps we need a module for array functions?
    def remove_duplicate_consecutive_tuples(tuple_list: list[tuple | npt.NDArray]) -> list[tuple]:
        """
        Remove duplicate consecutive tuples from a list.

        Parameters
        ----------
        tuple_list : list[tuple | npt.NDArray]
            List of tuples or numpy ndarrays to remove consecutive duplicates from.

        Returns
        -------
        list[Tuple]
            List of tuples with consecutive duplicates removed.

        Examples
        --------
        For the list of tuples [(1, 2), (1, 2), (1, 2), (2, 3), (2, 3), (3, 4)], this function will return
        [(1, 2), (2, 3), (3, 4)]
        """
        duplicates_removed = []
        for index, tup in enumerate(tuple_list):
            if index == 0 or not np.array_equal(tuple_list[index - 1], tup):
                duplicates_removed.append(tup)
        return np.array(duplicates_removed)

    def run_spline_trace(self) -> tuple[npt.NDArray, dict]:
        """
        Pipeline to run the splining smoothing and obtaining smoothing stats.

        Returns
        -------
        tuple[npt.NDArray, dict]
            Tuple of Nx2 smoothed trace coordinates, and smoothed trace statistics.
        """
        # fitted trace
        # fitted_trace = self.get_fitted_traces(self.ordered_trace, mol_is_circular)
        # splined trace
        splined_trace = self.get_splined_traces(self.mol_ordered_trace)
        # compile CL & E2E distance
        self.tracing_stats["contour_length"] = (
            measure_contour_length(splined_trace, self.mol_is_circular, self.pixel_to_nm_scaling) * 1e-9
        )
        self.tracing_stats["end_to_end_distance"] = (
            measure_end_to_end_distance(splined_trace, self.mol_is_circular, self.pixel_to_nm_scaling) * 1e-9
        )

        return splined_trace, self.tracing_stats


class windowTrace:
    """
    Obtain a smoothed trace of a molecule.

    Parameters
    ----------
    mol_ordered_tracing_data : dict
        Molecule ordered trace dictionary containing Nx2 ordered coords and molecule statistics.
    pixel_to_nm_scaling : float, optional
        The pixel to nm scaling factor, by default 1.
    rolling_window_size : np.float64, optional
        The length of the rolling window too average over, by default 6.0.
    rolling_window_resampling : bool, optional
        Whether to resample the rolling window, by default False.
    rolling_window_resample_regular_spatial_interval_nm : float, optional
        The regular spatial interval (nm) to resample the rolling window, by default 0.5.
    """

    def __init__(
        self,
        mol_ordered_tracing_data: dict,
        pixel_to_nm_scaling: float,
        rolling_window_size: float,
        rolling_window_resampling: bool = False,
        rolling_window_resample_regular_spatial_interval_nm: float = 0.5,
    ) -> None:
        """
        Initialise the windowTrace class.

        Parameters
        ----------
        mol_ordered_tracing_data : dict
            Molecule ordered trace dictionary containing Nx2 ordered coords and molecule statistics.
        pixel_to_nm_scaling : float, optional
            The pixel to nm scaling factor, by default 1.
        rolling_window_size : np.float64, optional
            The length of the rolling window too average over, by default 6.0.
        rolling_window_resampling : bool, optional
            Whether to resample the rolling window, by default False.
        rolling_window_resample_regular_spatial_interval_nm : float, optional
            The regular spatial interval (nm) to resample the rolling window, by default 0.5.
        """
        self.mol_ordered_trace = mol_ordered_tracing_data["ordered_coords"]
        self.mol_is_circular = mol_ordered_tracing_data["mol_stats"]["circular"]
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.rolling_window_size = rolling_window_size / 1e-9  # for nm scaling factor
        self.rolling_window_resampling = rolling_window_resampling
        self.rolling_window_resample_regular_spatial_interval_nm = (
            rolling_window_resample_regular_spatial_interval_nm / 1e-9
        )  # for nm scaling factor

        self.tracing_stats = {
            "contour_length": None,
            "end_to_end_distance": None,
        }

    @staticmethod
    def pool_trace_circular(
        pixel_trace: npt.NDArray[np.int32], rolling_window_size: np.float64 = 6.0, pixel_to_nm_scaling: float = 1
    ) -> npt.NDArray[np.float64]:
        """
        Smooth a pixelwise ordered trace of circular molecules via a sliding window.

        Parameters
        ----------
        pixel_trace : npt.NDArray[np.int32]
            Nx2 ordered trace coordinates.
        rolling_window_size : np.float64, optional
            The length of the rolling window too average over, by default 6.0.
        pixel_to_nm_scaling : float, optional
            The pixel to nm scaling factor, by default 1.

        Returns
        -------
        npt.NDArray[np.float64]
            MxN Smoothed ordered trace coordinates.
        """
        # Pool the trace points
        pooled_trace = []

        for i in range(len(pixel_trace)):
            binned_points = []
            current_length = 0
            j = 1

            # compile rolling window
            while current_length < rolling_window_size:
                current_index = i + j
                previous_index = i + j - 1
                while current_index >= len(pixel_trace):
                    current_index -= len(pixel_trace)
                while previous_index >= len(pixel_trace):
                    previous_index -= len(pixel_trace)
                current_length += (
                    np.linalg.norm(pixel_trace[current_index] - pixel_trace[previous_index]) * pixel_to_nm_scaling
                )
                binned_points.append(pixel_trace[current_index])
                j += 1

            # Get the mean of the binned points
            pooled_trace.append(np.mean(binned_points, axis=0))

        return np.array(pooled_trace)

    @staticmethod
    def pool_trace_linear(
        pixel_trace: npt.NDArray[np.int32],
        rolling_window_size: np.float64 = 6.0,
        pixel_to_nm_scaling: float = 1,
    ) -> npt.NDArray[np.float64]:
        """
        Smooth a pixelwise ordered trace of linear molecules via a sliding window.

        Parameters
        ----------
        pixel_trace : npt.NDArray[np.int32]
            Nx2 ordered trace coordinates.
        rolling_window_size : np.float64, optional
            The length of the rolling window too average over, by default 6.0.
        pixel_to_nm_scaling : float, optional
            The pixel to nm scaling factor, by default 1.

        Returns
        -------
        npt.NDArray[np.float64]
            MxN Smoothed ordered trace coordinates.
        """
        pooled_trace = [pixel_trace[0]]  # Add first coord as to not cut it off

        # Get average point for trace in rolling window
        for i in range(0, len(pixel_trace)):
            binned_points = []
            current_length = 0
            j = 0
            # Compile rolling window
            while current_length < rolling_window_size:
                current_index = i + j
                previous_index = i + j - 1
                if current_index >= len(pixel_trace):  # exit if exceeding the trace
                    break
                current_length += (
                    np.linalg.norm(pixel_trace[current_index] - pixel_trace[previous_index]) * pixel_to_nm_scaling
                )
                binned_points.append(pixel_trace[current_index])
                j += 1
            else:
                # Get the mean of the binned points
                pooled_trace.append(np.mean(binned_points, axis=0))

            # Exit if reached the end of the trace
            if current_index + 1 >= len(pixel_trace):
                break

        pooled_trace.append(pixel_trace[-1])  # Add last coord as to not cut it off

        # Check if the first two points are the same and remove the first point if they are
        # This can happen if the algorithm happens to add the first point naturally due to having a small
        # rolling window size.
        if np.array_equal(pooled_trace[0], pooled_trace[1]):
            pooled_trace.pop(0)

        # Check if the last two points are the same and remove the last point if they are
        # This can happen if the algorithm happens to add the last point naturally due to having a small
        # rolling window size.
        if np.array_equal(pooled_trace[-1], pooled_trace[-2]):
            pooled_trace.pop(-1)

        return np.array(pooled_trace)

    def run_window_trace(self) -> tuple[npt.NDArray, dict]:
        """
        Pipeline to run the rolling window smoothing and obtaining smoothing stats.

        Returns
        -------
        tuple[npt.NDArray, dict]
            Tuple of Nx2 smoothed trace coordinates, and smoothed trace statistics.
        """
        # fitted trace
        # fitted_trace = self.get_fitted_traces(self.ordered_trace, mol_is_circular)
        # splined trace
        if self.mol_is_circular:
            splined_trace = self.pool_trace_circular(
                self.mol_ordered_trace, self.rolling_window_size, self.pixel_to_nm_scaling
            )
            if self.rolling_window_resampling:
                splined_trace = resample_points_regular_interval(
                    points=splined_trace,
                    interval=self.rolling_window_resample_regular_spatial_interval_nm / self.pixel_to_nm_scaling,
                    circular=True,
                )
        else:
            splined_trace = self.pool_trace_linear(
                self.mol_ordered_trace, self.rolling_window_size, self.pixel_to_nm_scaling
            )
            if self.rolling_window_resampling:
                splined_trace = resample_points_regular_interval(
                    points=splined_trace,
                    interval=self.rolling_window_resample_regular_spatial_interval_nm / self.pixel_to_nm_scaling,
                    circular=False,
                )
        # compile CL & E2E distance
        self.tracing_stats["contour_length"] = (
            measure_contour_length(splined_trace, self.mol_is_circular, self.pixel_to_nm_scaling) * 1e-9
        )
        self.tracing_stats["end_to_end_distance"] = (
            measure_end_to_end_distance(splined_trace, self.mol_is_circular, self.pixel_to_nm_scaling) * 1e-9
        )

        return splined_trace, self.tracing_stats


def measure_contour_length(splined_trace: npt.NDArray, mol_is_circular: bool, pixel_to_nm_scaling: float) -> float:
    """
    Contour length for each of the splined traces accounting  for whether the molecule is circular or linear.

    Contour length units are nm.

    Parameters
    ----------
    splined_trace : npt.NDArray
        The splined trace.
    mol_is_circular : bool
        Whether the molecule is circular or not.
    pixel_to_nm_scaling : float
        Scaling factor from pixels to nanometres.

    Returns
    -------
    float
        Length of molecule in nanometres (nm).
    """
    if mol_is_circular:
        for num in range(len(splined_trace)):
            x1 = splined_trace[num - 1, 0]
            y1 = splined_trace[num - 1, 1]
            x2 = splined_trace[num, 0]
            y2 = splined_trace[num, 1]

            try:
                hypotenuse_array.append(math.hypot((x1 - x2), (y1 - y2)))
            except NameError:
                hypotenuse_array = [math.hypot((x1 - x2), (y1 - y2))]

        contour_length = np.sum(np.array(hypotenuse_array)) * pixel_to_nm_scaling
        del hypotenuse_array

    else:
        for num in range(len(splined_trace)):
            try:
                x1 = splined_trace[num, 0]
                y1 = splined_trace[num, 1]
                x2 = splined_trace[num + 1, 0]
                y2 = splined_trace[num + 1, 1]

                try:
                    hypotenuse_array.append(math.hypot((x1 - x2), (y1 - y2)))
                except NameError:
                    hypotenuse_array = [math.hypot((x1 - x2), (y1 - y2))]
            except IndexError:  # IndexError happens at last point in array
                contour_length = np.sum(np.array(hypotenuse_array)) * pixel_to_nm_scaling
                del hypotenuse_array
                break
    return contour_length


def measure_end_to_end_distance(splined_trace, mol_is_circular, pixel_to_nm_scaling: float):
    """
    Euclidean distance between the start and end of linear molecules.

    The hypotenuse is calculated between the start ([0,0], [0,1]) and end ([-1,0], [-1,1]) of linear
    molecules. If the molecule is circular then the distance is set to zero (0).

    Parameters
    ----------
    splined_trace : npt.NDArray
        The splined trace.
    mol_is_circular : bool
        Whether the molecule is circular or not.
    pixel_to_nm_scaling : float
        Scaling factor from pixels to nanometres.

    Returns
    -------
    float
        Length of molecule in nanometres (nm).
    """
    if not mol_is_circular:
        return (
            math.hypot((splined_trace[0, 0] - splined_trace[-1, 0]), (splined_trace[0, 1] - splined_trace[-1, 1]))
            * pixel_to_nm_scaling
        )
    return 0


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def splining_image(
    image: npt.NDArray,
    ordered_tracing_direction_data: dict,
    pixel_to_nm_scaling: float,
    filename: str,
    method: str,
    rolling_window_size: float,
    spline_step_size: float,
    spline_linear_smoothing: float,
    spline_circular_smoothing: float,
    spline_degree: int,
    rolling_window_resampling: bool = False,
    rolling_window_resample_regular_spatial_interval: float = 0.5,
) -> tuple[dict, pd.DataFrame, pd.DataFrame]:
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    """
    Obtain smoothed traces of pixel-wise ordered traces for molecules in an image.

    Parameters
    ----------
    image : npt.NDArray
        Whole image containing all molecules and grains.
    ordered_tracing_direction_data : dict
        Dictionary result from the ordered traces.
    pixel_to_nm_scaling : float
        Scaling factor from pixels to nanometres.
    filename : str
        Name of the image file.
    method : str
        Method of trace smoothing, options are 'splining' and 'rolling_window'.
    rolling_window_size : float
        Length in meters to average coordinates over in the rolling window.
    spline_step_size : float
        Step length in meters to use a coordinate for splining.
    spline_linear_smoothing : float
        Amount of linear spline smoothing.
    spline_circular_smoothing : float
        Amount of circular spline smoothing.
    spline_degree : int
        Degree of the spline. Cubic splines are recommended. Even values of k should be avoided especially with a
        small s-value.
    rolling_window_resampling : bool, optional
        Whether to resample the rolling window, by default False.
    rolling_window_resample_regular_spatial_interval : float, optional
        The regular spatial interval (nm) to resample the rolling window, by default 0.5.

    Returns
    -------
    tuple[dict, pd.DataFrame, pd.DataFrame]
        A spline data dictionary for all molecules, and a grainstats dataframe additions dataframe and molecule
        statistics dataframe.
    """
    grainstats_additions = {}
    molstats = {}
    all_splines_data = {}

    mol_count = 0
    for mol_trace_data in ordered_tracing_direction_data.values():
        mol_count += len(mol_trace_data)
    LOGGER.info(f"[{filename}] : Calculating Splining statistics for {mol_count} molecules...")

    # iterate through disordered_tracing_dict
    for grain_no, ordered_grain_data in ordered_tracing_direction_data.items():
        grain_trace_stats = {"total_contour_length": 0, "average_end_to_end_distance": 0}
        all_splines_data[grain_no] = {}
        mol_no = None
        for mol_no, mol_trace_data in ordered_grain_data.items():
            try:
                LOGGER.debug(f"[{filename}] : Splining {grain_no} - {mol_no}")
                # check if want to do nodestats tracing or not
                if method == "rolling_window":
                    splined_data, tracing_stats = windowTrace(
                        mol_ordered_tracing_data=mol_trace_data,
                        pixel_to_nm_scaling=pixel_to_nm_scaling,
                        rolling_window_size=rolling_window_size,
                        rolling_window_resampling=rolling_window_resampling,
                        rolling_window_resample_regular_spatial_interval_nm=rolling_window_resample_regular_spatial_interval,
                    ).run_window_trace()

                # if not doing nodestats ordering, do original TS ordering
                else:  # method == "spline":
                    splined_data, tracing_stats = splineTrace(
                        image=image,
                        mol_ordered_tracing_data=mol_trace_data,
                        pixel_to_nm_scaling=pixel_to_nm_scaling,
                        spline_step_size=spline_step_size,
                        spline_linear_smoothing=spline_linear_smoothing,
                        spline_circular_smoothing=spline_circular_smoothing,
                        spline_degree=spline_degree,
                    ).run_spline_trace()

                # get combined stats for the grains
                grain_trace_stats["total_contour_length"] += tracing_stats["contour_length"]
                grain_trace_stats["average_end_to_end_distance"] += tracing_stats["end_to_end_distance"]

                # get individual mol stats
                all_splines_data[grain_no][mol_no] = {
                    "spline_coords": splined_data,
                    "bbox": mol_trace_data["bbox"],
                    "tracing_stats": tracing_stats,
                }
                molstats[grain_no.split("_")[-1] + "_" + mol_no.split("_")[-1]] = {
                    "image": filename,
                    "grain_number": int(grain_no.split("_")[-1]),
                    "molecule_number": int(mol_no.split("_")[-1]),
                }
                molstats[grain_no.split("_")[-1] + "_" + mol_no.split("_")[-1]].update(tracing_stats)
                LOGGER.debug(f"[{filename}] : Finished splining {grain_no} - {mol_no}")

            except Exception as e:  # pylint: disable=broad-exception-caught
                LOGGER.error(
                    f"[{filename}] : Splining for {grain_no} failed. Consider raising an issue on GitHub. Error: ",
                    exc_info=e,
                )
                all_splines_data[grain_no] = {}

        if mol_no is None:
            LOGGER.warning(f"[{filename}] : No molecules found for grain {grain_no}")
        else:
            # average the e2e dists -> mol_no should always be in the grain dict
            grain_trace_stats["average_end_to_end_distance"] /= len(ordered_grain_data)

        # compile metrics
        grainstats_additions[grain_no] = {
            "image": filename,
            "grain_number": int(grain_no.split("_")[-1]),
        }
        grainstats_additions[grain_no].update(grain_trace_stats)

    # convert grainstats metrics to dataframe
    splining_stats_df = pd.DataFrame.from_dict(grainstats_additions, orient="index")
    molstats_df = pd.DataFrame.from_dict(molstats, orient="index")
    molstats_df.reset_index(drop=True, inplace=True)
    return all_splines_data, splining_stats_df, molstats_df


def interpolate_between_two_points_distance(
    point1: npt.NDArray[np.float32], point2: npt.NDArray[np.float32], distance: np.float32
) -> npt.NDArray[np.float32]:
    """
    Interpolate between two points to create a new point at a set distance between the two.

    Parameters
    ----------
    point1 : npt.NDArray[np.float32]
        The first point.
    point2 : npt.NDArray[np.float32]
        The second point.
    distance : np.float32
        The distance to interpolate between the two points.

    Returns
    -------
    npt.NDArray[np.float32]
        The new point at the specified distance between the two points.
    """
    distance_between_points = np.linalg.norm(point2 - point1)
    assert (
        distance_between_points > distance
    ), f"distance between points is less than the desired interval: {distance_between_points} < {distance}"
    proportion = distance / distance_between_points
    return point1 + proportion * (point2 - point1)


def resample_points_regular_interval(points: npt.NDArray, interval: float, circular: bool) -> npt.NDArray:
    """
    Resample a set of points to be at regular spatial intervals.

    Note: This is NOT intended to be pure interpolation, as interpolated points would not produce uniformly spaced
    points in cartesian space.

    Parameters
    ----------
    points : npt.NDArray
        The points to resample.
    interval : float
        The distance that all returned points should be apart.
    circular : bool
        If True, the first and last points will be connected to form a closed loop. If False, the first and last points
        will not be connected.

    Returns
    -------
    npt.NDArray
        The resampled points, evenly spaced at the specified interval.
    """
    if circular:
        points = np.concatenate((points, points[0:1]), axis=0)

    resampled_points = []
    resampled_points.append(points[0])
    current_point_index = 1
    while True:
        current_point = resampled_points[-1]
        next_original_point = points[current_point_index]
        distance_to_next_splined_point = np.linalg.norm(next_original_point - current_point)
        # if the distance to the next splined point is less than the interval, then skip to the next point
        if distance_to_next_splined_point < interval:
            current_point_index += 1
            if current_point_index >= len(points):
                break
            continue
        new_interpolated_point = interpolate_between_two_points_distance(
            point1=current_point, point2=next_original_point, distance=interval
        )
        resampled_points.append(new_interpolated_point)

    # if the first and last points are less than 0.5 * the interval apart, then remove the last point
    if np.linalg.norm(resampled_points[0] - resampled_points[-1]) < 0.5 * interval:
        resampled_points = resampled_points[:-1]

    return np.array(resampled_points)



"""Functions and tools for working with configuration files."""

import logging
from argparse import Namespace
from collections.abc import MutableMapping
from datetime import datetime
from pathlib import Path
from pkgutil import get_data
from pprint import pformat
from typing import TypeVar

import yaml

from topostats import CONFIG_DOCUMENTATION_REFERENCE
from topostats.io import read_yaml
from topostats.logs.logs import LOGGER_NAME
from topostats.utils import convert_path

MutableMappingType = TypeVar("MutableMappingType", bound="MutableMapping")

LOGGER = logging.getLogger(LOGGER_NAME)


def reconcile_config_args(args: Namespace | None) -> dict:
    """
    Reconcile command line arguments with the default configuration.

    Command line arguments take precedence over the default configuration. If a partial configuration file is specified
    (with '-c' or '--config-file') the defaults are over-ridden by these values (internally the configuration
    dictionary is updated with these values). Any other command line arguments take precedence over both the default
    and those supplied in a configuration file (again the dictionary is updated).

    The final configuration is validated before processing begins.

    Parameters
    ----------
    args : Namespace
        Command line arguments passed into TopoStats.

    Returns
    -------
    dict
        The configuration dictionary.
    """
    update_module(args=args)
    default_config = get_data(package=args.module, resource="default_config.yaml")
    default_config = yaml.full_load(default_config)
    if args is not None:
        if args.config_file is not None:
            config = read_yaml(str(args.config_file))
            # Merge the loaded config with the default config to fill in any defaults that are missing
            # Make sure to prioritise the loaded config, so it overrides the default
            config = merge_mappings(map1=default_config, map2=config)
        else:
            # If no config file is provided, use the default config
            config = default_config
    else:
        # If no args are provided, use the default config
        config = default_config

    # Override the config with command line arguments passed in, eg --output_dir ./output/
    if args is not None:
        config = update_config(config, args)

    return config


def update_module(
    args: Namespace,
    topostats_modules: tuple = (
        "bruker-rename",
        "create-config",
        "curvature",
        "disordered_tracing",
        "filter",
        "grains",
        "grainstats",
        "nodestats",
        "ordered_tracing",
        "process",
        "splining",
    ),
) -> None:
    """
    Update the `args.module` argument if processing TopoStats objects.

    This function allows the sub-parser command to map to the pipeline we wish to use. For now TopoStats has sub-parsers
    but it is the intention to introduce sub-sub-parsers for other modules such that eventually we invoke ``topostats```
    with a module argument followed by the step of processing.

    >>> topostats topostats filter
    >>> topostats topostats process
    >>> topostats afmslicer slice
    >>> topostats afmslicer process
    >>> topostats perovstats process

    Parameters
    ----------
    args : Namespace
        Default arguments that need parsing and updating.
    topostats_modules : tuple
        List of module names that are unique to TopoStats.
    """
    if args.module in topostats_modules:
        args.module = "topostats"


def merge_mappings(map1: MutableMappingType, map2: MutableMappingType) -> MutableMappingType:
    """
    Merge two mappings (dictionaries), with priority given to the second mapping.

    Note: Using a Mapping should make this robust to any mapping type, not just dictionaries. MutableMapping was needed
    as Mapping is not a mutable type, and this function needs to be able to change the dictionaries.

    Parameters
    ----------
    map1 : MutableMapping
        First mapping to merge, with secondary priority.
    map2 : MutableMapping
        Second mapping to merge, with primary priority.

    Returns
    -------
    dict
        Merged dictionary.
    """
    # Iterate over the second mapping
    for key, value in map2.items():
        # If the value is another mapping, then recurse
        if isinstance(value, MutableMapping):
            # If the key is not in the first mapping, add it as an empty dictionary before recursing
            map1[key] = merge_mappings(map1.get(key, {}), value)
        else:
            # Else simply add / override the key value pair
            map1[key] = value
    return map1


def write_config_with_comments(args: Namespace = None) -> None:  # noqa: C901
    """
    Write a sample configuration with in-line comments.

    This function is not designed to be used interactively but can be, just call it without any arguments and it will
    write a configuration to './config.yaml'.

    Parameters
    ----------
    args : Namespace
        A Namespace object parsed from argparse with values for 'filename'.
    """
    output_dir = Path("./") if args.output_dir is None else Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    logger_msg = "A sample configuration has been written to"
    # Update args.module
    update_module(args)
    if args.config == "default" or args.config is None:
        try:
            config = get_data(package=args.module, resource="default_config.yaml")
            filename = "default_config.yaml" if args.filename is None else args.filename
        except FileNotFoundError as exc:
            raise (
                FileNotFoundError(f"There is no configuration for module {args.module} called 'default_config.yaml'")
            ) from exc
    elif args.config == "simple":
        try:
            config = get_data(package=args.module, resource="simple_config.yaml")
            filename = "simple_config.yaml" if args.filename is None else args.filename
        except FileNotFoundError as exc:
            raise (
                FileNotFoundError(f"There is no configuration for module {args.module} called 'simple_config.yaml'")
            ) from exc
    elif args.config == "mplstyle":
        try:
            config = get_data(package=args.module, resource="topostats.mplstyle")
            filename = "topostats.mplstyle" if args.filename is None else args.filename
        except FileNotFoundError as exc:
            raise (
                FileNotFoundError(f"There is no configuration for module {args.module} called 'topostats.mplstyle'")
            ) from exc
    elif args.config == "var_to_label":
        try:
            config = get_data(package=args.module, resource="var_to_label.yaml")
            filename = "var_to_label.yaml" if args.filename is None else args.filename
        except FileNotFoundError as exc:
            raise (
                FileNotFoundError(f"There is no configuration for module {args.module} called 'var_to_label.yaml'")
            ) from exc
    else:
        valid_config = ["default", "simple", "mplstyle", "var_to_label"]
        raise ValueError(f"Invalid configuration file option, valid options are\n{valid_config}")

    if ".yaml" not in str(filename) and ".yml" not in str(filename) and ".mplstyle" not in str(filename):
        config_path = output_dir / f"{filename}.yaml"
    else:
        config_path = output_dir / filename

    try:
        with config_path.open("w", encoding="utf-8") as f:
            f.write(f"# Config file generated {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n")
            f.write(f"{CONFIG_DOCUMENTATION_REFERENCE}")
            f.write(config.decode("utf-8"))
    except AttributeError as e:
        raise e

    LOGGER.info(f"{logger_msg} : {str(config_path)}")


def update_config(config: dict, args: dict | Namespace) -> dict:
    """
    Update the configuration with any arguments.

    Parameters
    ----------
    config : dict
        Dictionary of configuration (typically read from YAML file specified with '-c/--config <filename>').
    args : Namespace
        Command line arguments.

    Returns
    -------
    dict
        Dictionary updated with command arguments.
    """
    args = vars(args) if isinstance(args, Namespace) else args

    config_keys = config.keys()
    for arg_key, arg_value in args.items():
        if isinstance(arg_value, dict):
            update_config(config, arg_value)
        else:
            if arg_key in config_keys and arg_value is not None:
                original_value = config[arg_key]
                config[arg_key] = arg_value
                LOGGER.debug(f"Updated config config[{arg_key}] : {original_value} > {arg_value} ")
    if "base_dir" in config.keys():
        config["base_dir"] = convert_path(config["base_dir"])
    if "output_dir" in config.keys():
        config["output_dir"] = convert_path(config["output_dir"])
    return config


def update_plotting_config(plotting_config: dict) -> dict:
    """
    Update the plotting config for each of the plots in plot_dict.

    Ensures that each entry has all the plotting configuration values that are needed.

    Parameters
    ----------
    plotting_config : dict
        Plotting configuration to be updated.

    Returns
    -------
    dict
        Updated plotting configuration.
    """
    main_config = plotting_config.copy()
    for opt in ["plot_dict", "run"]:
        main_config.pop(opt)
    LOGGER.debug(
        f"Main plotting options that need updating/adding to plotting dict :\n{pformat(main_config, indent=4)}"
    )
    for image, options in plotting_config["plot_dict"].items():
        main_config_temp = main_config.copy()
        LOGGER.debug(f"Dictionary for image : {image}")
        LOGGER.debug(f"{pformat(options, indent=4)}")
        # First update options with values that exist in main_config
        # We must however be careful not to update the colourmap for diagnostic traces
        if (
            not plotting_config["plot_dict"][image]["core_set"]
            and "mask_cmap" in plotting_config["plot_dict"][image].keys()
        ):
            main_config_temp.pop("mask_cmap")
        plotting_config["plot_dict"][image] = update_config(options, main_config_temp)
        LOGGER.debug(f"Updated values :\n{pformat(plotting_config['plot_dict'][image])}")
        # Then combine the remaining key/values we need from main_config that don't already exist
        for key_main, value_main in main_config_temp.items():
            if key_main not in plotting_config["plot_dict"][image]:
                plotting_config["plot_dict"][image][key_main] = value_main
        LOGGER.debug(f"After adding missing configuration options :\n{pformat(plotting_config['plot_dict'][image])}")
        # Make it so that binary images do not have the user-defined z-scale
        # applied, but non-binary images do.
        if plotting_config["plot_dict"][image]["image_type"] == "binary":
            plotting_config["plot_dict"][image]["zrange"] = [None, None]

    return plotting_config



"""Functions for measuring geometric properties of grains."""

import math

import networkx
import numpy as np
import numpy.typing as npt


def bounding_box_cartesian_points_float(
    points: npt.NDArray[np.number],
) -> tuple[np.float64, np.float64, np.float64, np.float64]:
    """
    Calculate the bounding box from a set of points.

    Parameters
    ----------
    points : npt.NDArray[np.number]
        Nx2 numpy array of points.

    Returns
    -------
    Tuple[np.float64, np.float64, np.float64, np.float64]
        Tuple of (min_x, min_y, max_x, max_y).

    Raises
    ------
    ValueError
        If the input array is not Nx2.
    """
    if points.shape[1] != 2:
        raise ValueError("Input array must be Nx2.")
    x_coords, y_coords = points[:, 0].astype(np.float64), points[:, 1].astype(np.float64)
    return (np.min(x_coords), np.min(y_coords), np.max(x_coords), np.max(y_coords))


def bounding_box_cartesian_points_integer(
    points: npt.NDArray[np.number],
) -> tuple[np.int32, np.int32, np.int32, np.int32]:
    """
    Calculate the bounding box from a set of points.

    Parameters
    ----------
    points : npt.NDArray[np.number]
        Nx2 numpy array of points.

    Returns
    -------
    Tuple[np.int32, np.int32, np.int32, np.int32]
        Tuple of (min_x, min_y, max_x, max_y).

    Raises
    ------
    ValueError
        If the input array is not Nx2.
    """
    if points.shape[1] != 2:
        raise ValueError("Input array must be Nx2.")
    x_coords, y_coords = points[:, 0].astype(np.int32), points[:, 1].astype(np.int32)
    return (np.min(x_coords), np.min(y_coords), np.max(x_coords), np.max(y_coords))


def do_points_in_arrays_touch(
    points1: npt.NDArray[np.int32], points2: npt.NDArray[np.int32]
) -> tuple[bool, npt.NDArray[np.int32] | None, npt.NDArray[np.int32] | None]:
    """
    Check if any points in two arrays are touching.

    Parameters
    ----------
    points1 : npt.NDArray[np.int32]
        Nx2 numpy array of points.
    points2 : npt.NDArray[np.int32]
        Mx2 numpy array of points.

    Returns
    -------
    tuple[bool, npt.NDArray[np.int32] | None, npt.NDArray[np.int32] | None]
        True if any points in the two arrays are touching, False otherwise, followed by the first touching point pair
        that was found. If no points are touching, the second and third elements of the tuple will be None.

    Raises
    ------
    ValueError
        If the input arrays are not Nx2 and Mx2.
    """
    if points1.shape[1] != 2 or points2.shape[1] != 2:
        raise ValueError("Input arrays must be Nx2 and Mx2.")

    for point1 in points1:
        for point2 in points2:
            diff = np.abs(point1 - point2)
            if np.all(diff <= 1):
                return (True, point1, point2)
    return (False, None, None)


# pylint: disable=too-many-locals
def calculate_shortest_branch_distances(
    nodes_with_branch_starting_coords: dict[int, list[npt.NDArray[np.int32]]],
    whole_skeleton_graph: networkx.classes.graph.Graph,
) -> tuple[npt.NDArray[np.number], npt.NDArray[np.int32], npt.NDArray[np.number]]:
    """
    Calculate the shortest distances between branches emanating from nodes.

    Parameters
    ----------
    nodes_with_branch_starting_coords : dict[int, list[npt.NDArray[np.int32]]]
        Dictionary where the key is the node number and the value is an Nx2 numpy array of the starting coordinates
        of its branches.
    whole_skeleton_graph : networkx.classes.graph.Graph
        Networkx graph representing the whole network.

    Returns
    -------
    Tuple[npt.NDArray[np.number], npt.NDArray[np.int32], npt.NDArray[np.int32]]
        - NxN numpy array of shortest distances between every node pair. Indexes of this array represent the nodes.
        Eg for a 3x3 matrix, there are 3 nodes being compared with each other.
        This matrix is diagonally symmetric and the diagonal values are 0 since a node is always 0 distance from itself.
        - NxNx2 numpy array of indexes of the best branches to connect between each node pair.
        Eg for node 1 and 3, the closest branches might be indexes 2 and 4, so the value at [1, 3] would be [2, 4].
        - NxNx2x2 numpy array of the coordinates of the branches to connect between each node pair.
        Eg for node 1 and 3, the closest branches might be at coordinates [2, 3] and [4, 5], so the
        value at [1, 3] would be [[2, 3], [4, 5]].
    """
    num_nodes = len(nodes_with_branch_starting_coords)
    shortest_node_distances = np.zeros((num_nodes, num_nodes), dtype=np.float64)
    # For storing the indexes of the branches that are the best candidate between two nodes.
    # Eg: [[[0, 0], [1, 2]], [[1, 2], [0, 0]]] means that node 0's branch 0 connects with node 1's branch 2.
    # Note that this matrix is symmetric about the diagonal as we double-iterate between all nodes.
    shortest_distances_branch_indexes = np.zeros((num_nodes, num_nodes, 2), dtype=np.int32)
    shortest_distances_branch_coordinates = np.zeros((num_nodes, num_nodes, 2, 2), dtype=object)

    # Iterate over the nodes twice to compare each combination of nodes. This double counts, so will create a symmetric
    # matrix about the diagonal.
    for node_index_i, (_node_i, node_branches_starts_coords_i) in enumerate(nodes_with_branch_starting_coords.items()):
        for node_index_j, (_node_j, node_branches_starts_coords_j) in enumerate(
            nodes_with_branch_starting_coords.items()
        ):
            # Don't compare the same node to itself
            if node_index_i == node_index_j:
                continue
            # Store the shortest distance as we iterate.
            shortest_distance = None
            # For storing the pair of branch indexes that are the best candidate between the two nodes.
            # Eg: (3, 2) means that node i's branch 3 connects with node j's branch 2.
            shortest_distance_branch_indexes: tuple[int, int] | None = None
            # Iteratively compare all branches from node1 to all branches from node2
            # to find the shortest distance between any two branches
            for branch_index_i, branch_start_i in enumerate(node_branches_starts_coords_i):
                for branch_index_j, branch_start_j in enumerate(node_branches_starts_coords_j):
                    shortest_path_length_between_branch_i_and_j = networkx.shortest_path_length(
                        whole_skeleton_graph, tuple(branch_start_i), tuple(branch_start_j)
                    )
                    if shortest_distance is None or shortest_path_length_between_branch_i_and_j < shortest_distance:
                        shortest_distance = shortest_path_length_between_branch_i_and_j
                        # Store the indexes of the branches that are the shortest distance apart for node i and node j
                        shortest_distance_branch_indexes = (branch_index_i, branch_index_j)

            # Store the shortest distance between the two nodes
            shortest_node_distances[node_index_i, node_index_j] = shortest_distance
            # Store the indexes of the branches that are the shortest distance apart for node i and node j.
            # Note this may be None as the nodes may not be connected?
            shortest_distances_branch_indexes[node_index_i, node_index_j] = shortest_distance_branch_indexes
            # Ensure that the nodes are connected before storing the coordinates of the branches starting coords
            if shortest_distance_branch_indexes is not None:
                # Add the coordinates of the branch pairs for each node-node combination. So for example, for
                # node 0 and node 1, branches starting [6, 1] and [6, 11]:
                # np.array([ [ [[0, 0][0, 0]] [[6 1][6 11]] ] [ [[6 11][6 1]] [[0, 0][0, 0]] ] ])
                # Where the square [0 0][0 0]s are for node 0 / node 0 and node 1 / node 1.
                # And [6 1][6 11] is for node 0 / node 1, indicating that branches start at [6, 1] and [6, 11].
                shortest_distances_branch_coordinates[node_index_i, node_index_j] = (
                    node_branches_starts_coords_i[shortest_distance_branch_indexes[0]],
                    node_branches_starts_coords_j[shortest_distance_branch_indexes[1]],
                )
            else:
                shortest_distances_branch_coordinates[node_index_i, node_index_j] = (None, None)

    return shortest_node_distances, shortest_distances_branch_indexes, shortest_distances_branch_coordinates


# pylint: disable=too-many-positional-arguments
def connect_best_matches(
    network_array_representation: npt.NDArray[np.int32],
    whole_skeleton_graph: networkx.classes.graph.Graph,
    match_indexes: npt.NDArray[np.int32],
    shortest_distances_between_nodes: npt.NDArray[np.number],
    shortest_distances_branch_indexes: npt.NDArray[np.int32],
    emanating_branch_starts_by_node: dict[int, list[npt.NDArray[np.int32]]],
    extend_distance: float = -1,
) -> npt.NDArray[np.int32]:
    """
    Connect the branches between node pairs that have been deemed to be best matches.

    Parameters
    ----------
    network_array_representation : npt.NDArray[np.int32]
        2D numpy array representing the network using integers to represent branches, nodes etc.
    whole_skeleton_graph : networkx.classes.graph.Graph
        Networkx graph representing the whole network.
    match_indexes : npt.NDArray[np.int32]
        Nx2 numpy array of indexes of the best matching nodes.
        Eg: np.array([[1, 0], [2, 3]]) means that the best matching nodes are node 1 and node 0, and node 2 and node 3.
    shortest_distances_between_nodes : npt.NDArray[np.number]
        NxN numpy array of shortest distances between every node pair.
        Index positions indicate which node it's referring to, so index 2, 3 will be the shortest distance
        between nodes 2 and 3.
        Values on the diagonal will be 0 because the shortest distance between a node and itself is 0.
        Eg: np.array([[0.0, 6.0], [6.0, 0.0]]) means that the shortest distance between node 0 and node 1 is 6.0.
    shortest_distances_branch_indexes : npt.NDArray[np.int32]
        NxNx2 numpy array of indexes of the branches to connect between the best matching nodes.
        Not entirely sure what it does so won't attempt to explain more to avoid confusion.
    emanating_branch_starts_by_node : dict[int, list[npt.NDArray[np.int32]]]
        Dictionary where the key is the node number and the value is an Nx2 numpy array of the starting coordinates
        of the branches emanating from that node. Rather self-explanatory.
        Eg:
        ```python
        {
            0: [np.array([6, 1]), np.array([7, 3]), np.array([8, 1])],
            1: [np.array([6, 11]), np.array([7, 9]), np.array([8, 11])],
        },
        ```.
    extend_distance : float
        The distance to extend the branches to connect. If the shortest distance between two nodes is less than or equal
        to this distance, the branches will be connected. If -1, the branches will be connected regardless of distance.

    Returns
    -------
    npt.NDArray[np.int32]
        2D numpy array representing the network using integers to represent branches, nodes etc.
    """
    for node_pair_index in match_indexes:
        # Fetch the shortest distance between two nodes
        shortest_distance = shortest_distances_between_nodes[node_pair_index[0], node_pair_index[1]]
        if shortest_distance <= extend_distance or extend_distance == -1:
            # Fetch the indexes of the branches to connect defined by the closest connecting
            # branchees of the given nodes
            indexes_of_branches_to_connect = shortest_distances_branch_indexes[node_pair_index[0], node_pair_index[1]]
            node_numbers = list(emanating_branch_starts_by_node.keys())
            # Grab the coordinate of the starting branch position of the branch to connect
            source = tuple(
                emanating_branch_starts_by_node[node_numbers[node_pair_index[0]]][indexes_of_branches_to_connect[0]]
            )
            # Grab the coordinate of the branch position to connect to
            target = tuple(
                emanating_branch_starts_by_node[node_numbers[node_pair_index[1]]][indexes_of_branches_to_connect[1]]
            )
            # Get the path between the two branches using networkx
            path = np.array(networkx.shortest_path(whole_skeleton_graph, source, target))
            # Set all the coordinates in the path to 3 in the network array representation
            network_array_representation[path[:, 0], path[:, 1]] = 3

    return network_array_representation


# pylint: disable=too-many-locals
def find_branches_for_nodes(
    network_array_representation: npt.NDArray[np.int32],
    labelled_nodes: npt.NDArray[np.int32],
    labelled_branches: npt.NDArray[np.int32],
) -> dict[int, list[npt.NDArray[np.int32]]]:
    """
    Locate branch starting positions for each node in a network.

    Parameters
    ----------
    network_array_representation : npt.NDArray[np.int32]
        2D numpy array representing the network using integers to represent branches, nodes etc.
    labelled_nodes : npt.NDArray[np.int32]
        2D numpy array representing the network using integers to represent nodes.
    labelled_branches : npt.NDArray[np.int32]
        2D numpy array representing the network using integers to represent branches.

    Returns
    -------
    dict[int, list[npt.NDArray[np.int32]]]
        Dictionary where the key is the node number and the value is an Nx2 numpy array of the starting coordinates
        of the branches emanating from that node.
    """
    # Dictionary to store emanating branches for each labelled node
    emanating_branch_starts_by_node = {}

    # Iterate over all the nodes in the labelled nodes image
    for node_num in range(1, labelled_nodes.max() + 1):
        num_branches = 0
        # makes lil box around node with 1 overflow
        bounding_box = bounding_box_cartesian_points_integer(np.argwhere(labelled_nodes == node_num))
        crop_left = bounding_box[0] - 1
        crop_right = bounding_box[2] + 2
        crop_top = bounding_box[1] - 1
        crop_bottom = bounding_box[3] + 2
        cropped_matrix = network_array_representation[crop_left:crop_right, crop_top:crop_bottom]
        # get coords of nodes and branches in box
        node_coords = np.argwhere(cropped_matrix == 3)
        branch_coords = np.argwhere(cropped_matrix == 1)
        # iterate through node coords to see which are within 8 dirs
        for node_coord in node_coords:
            for branch_coord in branch_coords:
                distance = math.dist(node_coord, branch_coord)
                if distance <= math.sqrt(2):
                    num_branches = num_branches + 1

        # All nodes with even branches are considered to be complete as they have one
        # strand going in for each coming out. This assumes that no strands naturally terminate at nodes.

        # find the branch start point of odd branched nodes
        if num_branches % 2 == 1:
            emanating_branches: list[npt.NDArray[np.int32]] = (
                []
            )  # List to store emanating branches for the current label
            for branch in range(1, labelled_branches.max() + 1):
                # technically using labelled_branches when there's an end loop will only cause one
                #   of the end loop coords to be captured. This shopuldn't matter as the other
                #   label after the crossing should be closer to another node.
                # The touching_point_1 and touching_point_2 can be None since the function returns None for both
                # if no points touch.
                touching, touching_point_1, _touching_point_2 = do_points_in_arrays_touch(
                    np.argwhere(labelled_branches == branch),
                    np.argwhere(labelled_nodes == node_num),
                )
                if touching:
                    assert touching_point_1 is not None
                    # Above required for mypy to ensure that there are no Nones
                    # in the list, to prevent the return type being npt.NDArray[np.int32 | None]
                    emanating_branches.append(touching_point_1)

            assert len(emanating_branches) > 0, f"No branches found for node {node_num}"
            emanating_branch_starts_by_node[node_num - 1] = (
                emanating_branches  # Store emanating branches for this label
            )

    return emanating_branch_starts_by_node



"""Utilities."""

import logging
from pathlib import Path

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.ndimage import convolve

from topostats.logs.logs import LOGGER_NAME
from topostats.thresholds import threshold

LOGGER = logging.getLogger(LOGGER_NAME)


COLUMN_SETS = {
    "grainstats": (
        "image",
        "basename",
        "grain_number",
        "class_number",
        "class_name",
        "subgrain_number",
        "area",
        "area_cartesian_bbox",
        "aspect_ratio",
        "bending_angle",
        "centre_x",
        "centre_y",
        "circular",
        "contour_length",
        "end_to_end_distance",
        "height_max",
        "height_mean",
        "height_median",
        "height_min",
        "max_feret",
        "min_feret",
        "radius_max",
        "radius_mean",
        "radius_median",
        "radius_min",
        "smallest_bounding_area",
        "smallest_bounding_length",
        "smallest_bounding_width",
        "threshold",
        "volume",
    ),
    "disordered_tracing_statistics": (
        "image",
        "basename",
        "threshold",
        "grain_number",
        "index",
        "branch_distance",
        "branch_type",
        "connected_segments",
        "mean_pixel_value",
        "stdev_pixel_value",
        "min_value",
        "median_value",
        "middle_value",
    ),
    "mol_statistics": (
        "image",
        "threshold",
        "basename",
        "grain_number",
        "molecule_number",
        "circular",
        "topology",
        "processing",
    ),
}


def convert_path(path: str | Path) -> Path:
    """
    Ensure path is Path object.

    Parameters
    ----------
    path : str | Path
        Path to be converted.

    Returns
    -------
    Path
        Pathlib object of path.
    """
    return Path().cwd() if path == "./" else Path(path).expanduser()


def _get_mask(image: npt.NDArray, thresh: float, threshold_direction: str, img_name: str = None) -> npt.NDArray:
    """
    Calculate a mask for pixels that exceed the threshold.

    Parameters
    ----------
    image : np.array
        Numpy array representing image.
    thresh : float
        A float representing the threshold.
    threshold_direction : str
        A string representing the direction that should be thresholded. ("above", "below").
    img_name : str
        Name of image being processed.

    Returns
    -------
    npt.NDArray
        Numpy array of image with objects coloured.
    """
    if threshold_direction == "above":
        LOGGER.debug(f"[{img_name}] : Masking (above) Threshold: {thresh}")
        return image > thresh
    LOGGER.debug(f"[{img_name}] : Masking (below) Threshold: {thresh}")
    return image < thresh
    # LOGGER.fatal(f"[{img_name}] : Threshold direction invalid: {threshold_direction}")


def get_mask(image: npt.NDArray, thresholds: dict, img_name: str = None) -> npt.NDArray:
    """
    Mask data that should not be included in flattening.

    Parameters
    ----------
    image : npt.NDArray
        2D Numpy array of the image to have a mask derived for.
    thresholds : dict
        Dictionary of thresholds, at a bare minimum must have key 'below' with an associated value, second key is
        to have an 'above' threshold.
    img_name : str
        Image name that is being masked.

    Returns
    -------
    npt.NDArray
        2D Numpy boolean array of points to mask.
    """
    # Both thresholds are applicable
    if "below" in thresholds and "above" in thresholds:
        mask_above = _get_mask(image, thresh=thresholds["above"], threshold_direction="above", img_name=img_name)
        mask_below = _get_mask(image, thresh=thresholds["below"], threshold_direction="below", img_name=img_name)
        # Masks are combined to remove both the extreme high and extreme low data points.
        return mask_above + mask_below
    # Only below threshold is applicable
    if "below" in thresholds:
        return _get_mask(image, thresh=thresholds["below"], threshold_direction="below", img_name=img_name)
    # Only above threshold is applicable
    return _get_mask(image, thresh=thresholds["above"], threshold_direction="above", img_name=img_name)


# pylint: disable=too-many-branches
def get_thresholds(  # noqa: C901
    image: npt.NDArray,
    threshold_method: str,
    otsu_threshold_multiplier: float | None = None,
    threshold_std_dev: dict[str, list] | None = None,
    absolute: dict[str, list] | None = None,
) -> dict[str, list[float]]:
    """
    Obtain thresholds for masking data points.

    Parameters
    ----------
    image : npt.NDArray
        2D Numpy array of image to be masked.
    threshold_method : str
        Method for thresholding, 'otsu', 'std_dev' or 'absolute' are valid options.
    otsu_threshold_multiplier : float
        Scaling value for Otsu threshold.
    threshold_std_dev : dict
        Dict of above and below thresholds for the standard deviation method.
    absolute : tuple
        Dict of below and above thresholds.

    Returns
    -------
    dict[str, list[float]]
        Dictionary of thresholds, contains keys 'below' and optionally 'above'.
    """
    thresholds: dict[str, list[float]] = {}
    if threshold_method == "otsu":
        assert (
            otsu_threshold_multiplier is not None
        ), "Otsu threshold multiplier must be provided when using 'otsu' thresholding method."
        thresholds["above"] = [threshold(image, method="otsu", otsu_threshold_multiplier=otsu_threshold_multiplier)]
    elif threshold_method == "std_dev":
        assert (
            threshold_std_dev is not None
        ), "Standard deviation thresholds must be provided when using 'std_dev' thresholding method."
        if threshold_std_dev["below"] is not None:
            thresholds_std_dev_below = []
            for threshold_std_dev_value in threshold_std_dev["below"]:
                thresholds_std_dev_below.append(
                    threshold(image, method="mean") - threshold_std_dev_value * np.nanstd(image)
                )
            thresholds["below"] = thresholds_std_dev_below
        if threshold_std_dev["above"] is not None:
            thresholds_std_dev_above = []
            for threshold_std_dev_value in threshold_std_dev["above"]:
                thresholds_std_dev_above.append(
                    threshold(image, method="mean") + threshold_std_dev_value * np.nanstd(image)
                )
            thresholds["above"] = thresholds_std_dev_above
    elif threshold_method == "absolute":
        assert absolute is not None, "Absolute thresholds must be provided when using 'absolute' thresholding method."
        if absolute["below"] is not None:
            thresolds_absolute_below = []
            for threshold_absolute_value in absolute["below"]:
                thresolds_absolute_below.append(threshold_absolute_value)
            thresholds["below"] = thresolds_absolute_below
        if absolute["above"] is not None:
            thresolds_absolute_above = []
            for threshold_absolute_value in absolute["above"]:
                thresolds_absolute_above.append(threshold_absolute_value)
            thresholds["above"] = thresolds_absolute_above
    else:
        if not isinstance(threshold_method, str):
            raise TypeError(
                f"threshold_method ({threshold_method}) should be a string. Valid values : 'otsu' 'std_dev' 'absolute'"
            )
        raise ValueError(
            f"threshold_method ({threshold_method}) is invalid. Valid values : 'otsu' 'std_dev' 'absolute'"
        )
    return thresholds


def create_empty_dataframe(column_set: str = "grainstats") -> pd.DataFrame:
    """
    Create an empty data frame for returning when no results are found.

    Parameters
    ----------
    column_set : str
        The name of the set of columns for the empty dataframe.

    Returns
    -------
    pd.DataFrame
        Empty Pandas DataFrame.
    """
    return pd.DataFrame(columns=COLUMN_SETS[column_set])


def bound_padded_coordinates_to_image(coordinates: npt.NDArray, padding: int, image_shape: tuple) -> tuple:
    """
    Ensure the padding of coordinates points does not fall outside of the image shape.

    This function is primarily used in the dnaTrace.get_fitted_traces() method which aims to adjust the points of a
    skeleton to sit on the highest points of a traced molecule. In order to do so it takes the ordered skeleton, which
    may not lie on the highest points as it is generated from a binary mask that is unaware of the heights, and then
    defines a padded boundary of 3nm profile perpendicular to the backbone of the DNA (which at this point is the
    skeleton based on a mask). Each point along the skeleton therefore needs padding by a minimum of 2 pixels (in this
    case each pixel equates to a cell in a NumPy array). If a point is within 2 pixels (i.e. 2 cells) of the border then
    we can not pad beyond this region, we have to stop at the edge of the image and so the coordinates is adjusted such
    that the padding will lie on the edge of the image/array.

    Parameters
    ----------
    coordinates : npt.NDArray
        Coordinates of a point on the mask based skeleton.
    padding : int
        Number of pixels/cells to pad around the point.
    image_shape : tuple
        The shape of the original image from which the pixel is obtained.

    Returns
    -------
    tuple
        Returns a tuple of coordinates that ensure that when the point is padded by the noted padding width in
        subsequent calculations it will not be outside of the image shape.
    """
    # Calculate the maximum row and column indexes
    max_row = image_shape[0] - 1
    max_col = image_shape[1] - 1
    row_coord, col_coord = coordinates

    def check(coord: npt.NDArray, max_val: int, padding: int) -> npt.NDArray:
        """
        Check coordinates are within the bounds of the padding.

        Parameters
        ----------
        coord : npt.NDArray
            Coordinates (length = 2).
        max_val : int
            Maximum width in the dimension being checked (max_row or max_col).
        padding : int
            Padding used in the image.

        Returns
        -------
        npt.NDArray
            Coordinates adjusted for padding.
        """
        if coord - padding < 0:
            coord = padding
        elif coord + padding > max_val:
            coord = max_val - padding
        return coord

    return check(row_coord, max_row, padding), check(col_coord, max_col, padding)


def convolve_skeleton(skeleton: npt.NDArray) -> npt.NDArray:
    """
    Convolve skeleton with a 3x3 kernel.

    This produces an array where the branches of the skeleton are denoted with '1', endpoints are denoted as '2', and
    pixels at nodes as '3'.

    Parameters
    ----------
    skeleton : npt.NDArray
        Single pixel thick binary trace(s) within an array.

    Returns
    -------
    npt.NDArray
        The skeleton (=1) with endpoints (=2), and crossings (=3) highlighted.
    """
    conv = convolve(skeleton.astype(np.int32), np.ones((3, 3)))
    conv[skeleton == 0] = 0  # remove non-skeleton points
    conv[conv == 3] = 1  # skelly = 1
    conv[conv > 3] = 3  # nodes = 3
    return conv


class ResolutionError(Exception):
    """Raised when the image resolution is too small for accuurate tracing."""

    pass  # pylint: disable=unnecessary-pass


def coords_2_img(coords, image, ordered=False) -> np.ndarray:
    """
    Convert coordinates to a binary image.

    Parameters
    ----------
    coords : np.ndarray
        An array of 2xN integer coordinates.
    image : np.ndarray
        An MxL array to assign the above coordinates onto.
    ordered : bool, optional
        If True, increments the value of each coord to show order.

    Returns
    -------
    np.ndarray
        An array the same shape as 'image' with the coordinates highlighted.
    """
    comb = np.zeros_like(image)
    if ordered:
        comb[coords[:, 0].astype(np.int32), coords[:, 1].astype(np.int32)] = np.arange(1, len(coords) + 1)
    else:
        coords = coords[
            (coords[:, 0] < image.shape[0]) & (coords[:, 1] < image.shape[1]) & (coords[:, 0] > 0) & (coords[:, 1] > 0)
        ]
        comb[np.floor(coords[:, 0]).astype(np.int32), np.floor(coords[:, 1]).astype(np.int32)] = 1
    return comb



"""Standardise logging."""

import logging
import sys
from datetime import datetime
from pathlib import Path

# pylint: disable=assignment-from-no-return

start = datetime.now()
LOG_INFO_FORMATTER = logging.Formatter(
    fmt="[%(asctime)s] [%(levelname)-8s] [%(name)s] %(message)s", datefmt="%a, %d %b %Y %H:%M:%S"
)
LOG_ERROR_FORMATTER = logging.Formatter(
    fmt="[%(asctime)s] [%(levelname)-8s] [%(name)s] [%(filename)s] [%(lineno)s] %(message)s",
    datefmt="%a, %d %b %Y %H:%M:%S",
)

LOGGER_NAME = "topostats"


def setup_logger(log_name: str = LOGGER_NAME) -> logging.Logger:
    """
    Logger setup.

    The logger for the module is initialised when the module is loaded (as this functions is called from
    __init__.py). This creates two stream handlers, one for general output and one for errors which are formatted
    differently (there is greater information in the error formatter). To use in modules import the 'LOGGER_NAME' and
    create a logger as shown in the Examples, it will inherit the formatting and direction of messages to the correct
    stream.

    Parameters
    ----------
    log_name : str
        Name under which logging information occurs.

    Returns
    -------
    logging.Logger
        Logger object.

    Examples
    --------
    To use the logger in (sub-)modules have the following.

        import logging
        from topostats.logs.logs import LOGGER_NAME

        LOGGER = logging.getLogger(LOGGER_NAME)

        LOGGER.info('This is a log message.')
    """
    out_stream_handler = logging.StreamHandler(sys.stdout)
    out_stream_handler.setLevel(logging.DEBUG)
    out_stream_handler.setFormatter(LOG_INFO_FORMATTER)

    err_stream_handler = logging.StreamHandler(sys.stderr)
    err_stream_handler.setLevel(logging.ERROR)
    err_stream_handler.setFormatter(LOG_ERROR_FORMATTER)

    file_handler = logging.FileHandler(Path().cwd().stem + f"-{start.strftime('%Y-%m-%d-%H-%M-%S')}.log")
    file_handler.setFormatter(LOG_ERROR_FORMATTER)

    logger = logging.getLogger(log_name)
    logger.setLevel(logging.INFO)
    logger.propagate = True
    if not logger.handlers:
        logger.addHandler(out_stream_handler)
        logger.addHandler(err_stream_handler)
        logger.addHandler(file_handler)

    return logger



"""Custom Bruker Nanoscope colorscale."""

import logging

import matplotlib as mpl
import matplotlib.cm
import numpy as np
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


class Colormap:
    """
    Class for setting the Colormap.

    Parameters
    ----------
    name : str
        Name of colormap to use.
    """

    def __init__(self, name: str = "nanoscope"):
        """
        Initialise the class.

        Parameters
        ----------
        name : str
            Name of colormap to use.
        """
        self.name = name
        self.cmap = None
        self.set_cmap(self.name)

    def __str__(self) -> str:
        """
        Return string representation of object.

        Returns
        -------
        str
            String detailing the colormap.
        """
        return f"TopoStats Colormap: {self.name}"

    def set_cmap(self, name: str) -> None:
        """
        Set the ColorMap.

        Parameters
        ----------
        name : str
            Name of the colormap to return.
        """
        if name.lower() == "nanoscope":
            self.cmap = self.nanoscope()
        elif name.lower() == "gwyddion":
            self.cmap = self.gwyddion()
        elif name.lower() == "blue":
            self.cmap = self.blue()
        elif name.lower() == "blue_purple_green":
            self.cmap = self.blue_purple_green()
        else:
            # Get one of the matplotlib colormaps
            self.cmap = mpl.colormaps[name]
        LOGGER.debug(f"[theme] Colormap set to : {name}")

    def get_cmap(self) -> matplotlib.cm:
        """
        Return the matplotlib.cm colormap object.

        Returns
        -------
        matplotlib.cm
            Matplotlib Color map object.
        """
        return self.cmap

    @staticmethod
    def nanoscope() -> LinearSegmentedColormap:
        """
        Matplotlib compatible colormap that replicates the Bruker Nanoscope colorscale.

        The colormap is implemented in Gwyddion's GwyGradient via 'Nanoscope.txt'.

        Returns
        -------
        LinearSegmentedColormap
            MatplotLib LinearSegmentedColourmap that replicates Bruker Nanoscope colorscale.
        """
        cdict = {
            "red": (
                (0.0, 0.0, 0.0),
                (0.124464, 0.0, 0.0),
                (0.236052, 0.0670103, 0.0670103),
                (0.371245, 0.253338, 0.253338),
                (0.472103, 0.392344, 0.392344),
                (0.611588, 0.584587, 0.584587),
                (0.708155, 0.717678, 0.717678),
                (0.714052, 0.725806, 0.725806),
                (0.890558, 0.969072, 0.969072),
                (0.933476, 0.987464, 0.987464),
                (0.944709, 0.992278, 0.992278),
                (0.965682, 0.995207, 0.995207),
                (0.971401, 0.996006, 0.996006),
                (1, 1, 1),
            ),
            "green": (
                (0.0, 0.0, 0.0),
                (0.124464, 0.0, 0.0),
                (0.236052, 0.0, 0.0),
                (0.371245, 0.0, 0.0),
                (0.472103, 0.0721649, 0.0721649),
                (0.611588, 0.334114, 0.334114),
                (0.708155, 0.515464, 0.515464),
                (0.714052, 0.527471, 0.527471),
                (0.890558, 0.886843, 0.886843),
                (0.933476, 0.974227, 0.974227),
                (0.944709, 0.980523, 0.980523),
                (0.965682, 0.992278, 0.992278),
                (0.971401, 0.993565, 0.993565),
                (1, 1, 1),
            ),
            "blue": (
                (0.0, 0.0, 0.0),
                (0.124464, 0.0, 0.0),
                (0.236052, 0.0, 0.0),
                (0.371245, 0.0, 0.0),
                (0.472103, 0.0, 0.0),
                (0.611588, 0.0, 0.0),
                (0.708155, 0.252575, 0.252575),
                (0.714052, 0.268, 0.268),
                (0.890558, 0.76343, 0.76343),
                (0.933476, 0.883897, 0.883897),
                (0.944709, 0.915426, 0.915426),
                (0.965682, 0.974293, 0.974293),
                (0.971401, 0.990347, 0.990347),
                (1, 1, 1),
            ),
        }

        return LinearSegmentedColormap("nanoscope", cdict)

    @staticmethod
    def gwyddion() -> LinearSegmentedColormap:
        """
        Set RGBA colour map for the Gwyddion.net colour gradient.

        Returns
        -------
        LinearSegmentedColormap
            The 'gwyddion' colormap.
        """
        N = 4  # Number of values
        vals = np.ones((N, 4))  # Initialise the array to be full of 1.0
        vals[0] = [0.0, 0.0, 0.0, 1]
        vals[1] = [168 / 256, 40 / 256, 15 / 256, 1.0]
        vals[2] = [243 / 256, 194 / 256, 93 / 256, 1.0]
        vals[3] = [1.0, 1.0, 1.0, 1.0]

        return LinearSegmentedColormap.from_list("gwyddion", vals, N=256)

    @staticmethod
    def blue() -> ListedColormap:
        """
        Set RGBA colour map of just the colour blue.

        Returns
        -------
        ListedColormap
            The 'blue' colormap.
        """
        return ListedColormap([[32 / 256, 226 / 256, 205 / 256]], "blue", N=256)

    @staticmethod
    def blue_purple_green() -> ListedColormap:
        """
        RGBA colour map of just the colour blue/purple/green.

        Returns
        -------
        ListedColormap
            The 'blue/purple/green' colormap.
        """
        return ListedColormap(
            [[0 / 256, 157 / 256, 229 / 256], [255 / 256, 100 / 256, 225 / 256], [0 / 256, 1, 139 / 256]],
            "blue_purple_green",
            N=3,
        )



"""Module for filtering 2D Numpy arrays."""

import logging

import numpy as np
import numpy.typing as npt
from scipy.optimize import curve_fit

# pylint: disable=no-name-in-module
from skimage.filters import gaussian

from topostats import scars
from topostats.logs.logs import LOGGER_NAME
from topostats.utils import get_mask, get_thresholds

LOGGER = logging.getLogger(LOGGER_NAME)

# noqa: PLR0913
# pylint: disable=fixme
# pylint: disable=broad-except
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-branches
# pylint: disable=dangerous-default-value


class Filters:
    """
    Class for filtering scans.

    Parameters
    ----------
    image : npt.NDArray
        The raw image from the Atomic Force Microscopy machine.
    filename : str
        The filename (used in logging only).
    pixel_to_nm_scaling : float
        Value for converting pixels to nanometers.
    row_alignment_quantile : float
        Quantile (0.0 to 1.0) to be used to determine the average background for the image below values may improve
        flattening of large features.
    threshold_method : str
        Method for thresholding, default 'otsu', valid options 'otsu', 'std_dev' and 'absolute'.
    otsu_threshold_multiplier : float
        Value for scaling the derived Otsu threshold.
    threshold_std_dev : dict
        If using the 'std_dev' threshold method. Dictionary that contains above and below threshold values for the
        number of standard deviations from the mean to threshold.
    threshold_absolute : dict
        If using the 'absolute' threshold method. Dictionary that contains above and below absolute threshold values
        for flattening.
    gaussian_size : float
        If using the 'absolute' threshold method. Dictionary that contains above and below absolute threshold values
        for flattening.
    gaussian_mode : str
        Method passed to 'skimage.filters.gaussian(mode = gaussian_mode)'.
    remove_scars : dict
        Dictionary containing configuration parameters for the scar removal function.
    """  # numpydoc: ignore=PR01

    def __init__(
        self,
        image: npt.NDArray,
        filename: str,
        pixel_to_nm_scaling: float,
        row_alignment_quantile: float = 0.5,
        threshold_method: str = "otsu",
        otsu_threshold_multiplier: float = 1.7,
        threshold_std_dev: dict | None = None,
        threshold_absolute: dict | None = None,
        gaussian_size: float = None,
        gaussian_mode: str = "nearest",
        remove_scars: dict = None,
    ):
        """
        Initialise the class.

        Parameters
        ----------
        image : npt.NDArray
            The raw image from the Atomic Force Microscopy machine.
        filename : str
            The filename (used in logging only).
        pixel_to_nm_scaling : float
            Value for converting pixels to nanometers.
        row_alignment_quantile : float
            Quantile (0.0 to 1.0) to be used to determine the average background for the image below values may improve
            flattening of large features.
        threshold_method : str
            Method for thresholding, default 'otsu', valid options 'otsu', 'std_dev' and 'absolute'.
        otsu_threshold_multiplier : float
            Value for scaling the derived Otsu threshold.
        threshold_std_dev : dict
            If using the 'std_dev' threshold method. Dictionary that contains above and below threshold values for the
            number of standard deviations from the mean to threshold.
        threshold_absolute : dict
            If using the 'absolute' threshold method. Dictionary that contains above and below absolute threshold values
            for flattening.
        gaussian_size : float
            If using the 'absolute' threshold method. Dictionary that contains above and below absolute threshold values
            for flattening.
        gaussian_mode : str
            Method passed to 'skimage.filters.gaussian(mode = gaussian_mode)'.
        remove_scars : dict
            Dictionary containing configuration parameters for the scar removal function.
        """
        self.filename = filename
        self.pixel_to_nm_scaling = pixel_to_nm_scaling
        self.gaussian_size = gaussian_size
        self.gaussian_mode = gaussian_mode
        self.row_alignment_quantile = row_alignment_quantile
        self.threshold_method = threshold_method
        self.otsu_threshold_multiplier = otsu_threshold_multiplier
        # Convert to lists since the thresholding function expects lists of thresholds but
        # we don't want to use more than one value for the filters step.
        if threshold_std_dev is None:
            threshold_std_dev = {"above": 1.0, "below": 1.0}
        else:
            self.threshold_std_dev = {
                "above": [threshold_std_dev["above"]],
                "below": [threshold_std_dev["below"]],
            }
        if threshold_absolute is None:
            threshold_absolute = {"above": 1.0, "below": 10.0}
        else:
            self.threshold_absolute = {
                "above": [threshold_absolute["above"]],
                "below": [threshold_absolute["below"]],
            }
        self.remove_scars_config = remove_scars
        self.images = {
            "pixels": image,
            "initial_median_flatten": None,
            "initial_tilt_removal": None,
            "initial_quadratic_removal": None,
            "initial_scar_removal": None,
            "initial_zero_average_background": None,
            "masked_median_flatten": None,
            "masked_tilt_removal": None,
            "masked_quadratic_removal": None,
            "secondary_scar_removal": None,
            "scar_mask": None,
            "mask": None,
            "final_zero_average_background": None,
            "gaussian_filtered": None,
        }
        self.thresholds = None
        self.medians = {"rows": None, "cols": None}
        self.results = {
            "diff": None,
            "median_row_height": None,
            "x_gradient": None,
            "y_gradient": None,
            "threshold": None,
        }

    def median_flatten(
        self, image: npt.NDArray, mask: npt.NDArray = None, row_alignment_quantile: float = 0.5
    ) -> npt.NDArray:
        """
        Flatten images using median differences.

        Flatten the rows of an image, aligning the rows and centering the median around zero. When used with a mask,
        this has the effect of centering the background data on zero.

        Note this function does not handle scars.

        Parameters
        ----------
        image : npt.NDArray
            2-D image of the data to align the rows of.
        mask : npt.NDArray
            Boolean array of points to mask (ignore).
        row_alignment_quantile : float
            Quantile (in the range 0.0 to 1.0) used for defining the average background.

        Returns
        -------
        npt.NDArray
            Copy of the input image with rows aligned.
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.debug(f"[{self.filename}] : Median flattening with mask")
        else:
            read_matrix = image
            LOGGER.debug(f"[{self.filename}] : Median flattening without mask")

        for row in range(image.shape[0]):
            # Get the median of the row
            m = np.nanquantile(read_matrix[row, :], row_alignment_quantile)
            if not np.isnan(m):
                image[row, :] -= m
            else:
                LOGGER.warning(
                    """f[{self.filename}] Large grain detected image can not be
processed, please refer to https://github.com/AFM-SPM/TopoStats/discussions for more information."""
                )

        return image

    def remove_tilt(self, image: npt.NDArray, mask: npt.NDArray = None) -> npt.NDArray:
        """
        Remove the planar tilt from an image (linear in 2D spaces).

        Uses a linear fit of the medians of the rows and columns to determine the linear slants in x and y directions
        and then subtracts the fit from the columns.

        Parameters
        ----------
        image : npt.NDArray
            2-D image of the data to remove the planar tilt from.
        mask : npt.NDArray
            Boolean array of points to mask (ignore).

        Returns
        -------
        npt.NDArray
            Numpy array of image with tilt removed.
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.debug(f"[{self.filename}] : Plane tilt removal with mask")
        else:
            read_matrix = image
            LOGGER.debug(f"[{self.filename}] : Plane tilt removal without mask")

        # Line of best fit
        # Calculate medians
        medians_x = [np.nanmedian(read_matrix[:, i]) for i in range(read_matrix.shape[1])]
        medians_y = [np.nanmedian(read_matrix[j, :]) for j in range(read_matrix.shape[0])]
        LOGGER.debug(f"[{self.filename}] [remove_tilt] medians_x   : {medians_x}")
        LOGGER.debug(f"[{self.filename}] [remove_tilt] medians_y   : {medians_y}")

        # Fit linear x
        px = np.polyfit(range(0, len(medians_x)), medians_x, 1)
        LOGGER.debug(f"[{self.filename}] : x-polyfit 1st order: {px}")
        py = np.polyfit(range(0, len(medians_y)), medians_y, 1)
        LOGGER.debug(f"[{self.filename}] : y-polyfit 1st order: {py}")

        if px[0] != 0:
            if not np.isnan(px[0]):
                LOGGER.debug(f"[{self.filename}] : Removing x plane tilt")
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= px[0] * (col)
            else:
                LOGGER.debug(f"[{self.filename}] : x gradient is nan, skipping plane tilt x removal")
        else:
            LOGGER.debug("[{self.filename}] : x gradient is zero, skipping plane tilt x removal")

        if py[0] != 0:
            if not np.isnan(py[0]):
                LOGGER.debug(f"[{self.filename}] : removing y plane tilt")
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= py[0] * (row)
            else:
                LOGGER.debug("[{self.filename}] : y gradient is nan, skipping plane tilt y removal")
        else:
            LOGGER.debug("[{self.filename}] : y gradient is zero, skipping plane tilt y removal")

        return image

    def remove_nonlinear_polynomial(self, image: npt.NDArray, mask: npt.NDArray | None = None) -> npt.NDArray:
        """
        Fit and remove a "saddle" shaped nonlinear polynomial from the image.

        "Saddles" with the form a + b * x * y - c * x - d * y from the supplied image. AFM images sometimes contain a
        "saddle" shape trend to their background, and so to remove them we fit a nonlinear polynomial of x and y and
        then subtract the fit from the image.

        If these trends are not removed, then the image will not flatten properly and will leave opposite diagonal
        corners raised or lowered.

        Parameters
        ----------
        image : npt.NDArray
            2-D numpy height-map array of floats with a polynomial trend to remove.
        mask : npt.NDArray, optional
            2-D Numpy boolean array used to mask any points in the image that are deemed not to be part of the
            height-map's background data.

        Returns
        -------
        npt.NDArray
            Image with the polynomial trend subtracted.
        """
        # Script has a lot of locals but I feel this is necessary for readability?
        # pylint: disable=too-many-locals

        # Define the polynomial function to fit to the image
        def model_func(x: float, y: float, a: float, b: float, c: float, d: float) -> float:
            """
            Polynomial function to fit to the image.

            Parameters
            ----------
            x : float
                X.
            y : float
                Y.
            a : float
                A.
            b : float
                B.
            c : float
                C.
            d : float
                D.

            Returns
            -------
            float
                Result of applying the polynomial a + (b * x * y) - (c * x) - (d * y).
            """
            return a + b * x * y - c * x - d * y

        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
        else:
            read_matrix = image

        # Construct a meshgrid of x and y points for fitting to the z heights
        xdata, ydata = np.meshgrid(np.arange(read_matrix.shape[1]), np.arange(read_matrix.shape[0]))
        zdata = read_matrix

        # Only use data that is not nan. Nans may be in the image from the
        # masked array. Curve fitting cannot handle nans.
        nan_mask = ~np.isnan(zdata)
        xdata_nans_removed = xdata[nan_mask]
        ydata_nans_removed = ydata[nan_mask]
        zdata_nans_removed = zdata[nan_mask]

        # Convert the z data to a 1D array
        zdata = zdata.ravel()
        zdata_nans_removed = zdata_nans_removed.ravel()

        # Stack the x, y meshgrid data after converting them to 1D
        xy_data_stacked = np.vstack((xdata_nans_removed.ravel(), ydata_nans_removed.ravel()))

        # Fit the model to the data
        # Note: pylint is flagging the tuple unpacking regarding an internal line of scipy.optimize._minpack_py : 910.
        # This isn't actually an issue though as the extended tuple output is only provided if the 'full_output' flag is
        # provided as a kwarg in curve_fit.
        popt, _pcov = curve_fit(  # pylint: disable=unbalanced-tuple-unpacking
            lambda x, a, b, c, d: model_func(x[0], x[1], a, b, c, d),
            xy_data_stacked,
            zdata_nans_removed,
        )

        # Unpack the optimised parameters
        a, b, c, d = popt
        LOGGER.debug(
            f"[{self.filename}] : Nonlinear polynomial removal optimal params: const: {a} xy: {b} x: {c} y: {d}"
        )

        # Use the optimised parameters to construct a prediction of the underlying surface
        z_pred = model_func(xdata, ydata, a, b, c, d)
        # Subtract the fitted nonlinear polynomial from the image
        image -= z_pred

        return image

    def remove_quadratic(self, image: npt.NDArray, mask: npt.NDArray = None) -> npt.NDArray:
        """
        Remove the quadratic bowing that can be seen in some large-scale AFM images.

        Use a simple quadratic fit on the medians of the columns of the image and then subtracts the calculated
        quadratic from the columns.

        Parameters
        ----------
        image : npt.NDArray
            2-D image of the data to remove the quadratic from.
        mask : npt.NDArray
            Boolean array of points to mask (ignore).

        Returns
        -------
        npt.NDArray
            Image with the quadratic bowing removed.
        """
        image = image.copy()
        if mask is not None:
            read_matrix = np.ma.masked_array(image, mask=mask, fill_value=np.nan).filled()
            LOGGER.debug(f"[{self.filename}] : Remove quadratic bow with mask")
        else:
            read_matrix = image
            LOGGER.debug(f"[{self.filename}] : Remove quadratic bow without mask")

        # Calculate medians
        medians_x = [np.nanmedian(read_matrix[:, i]) for i in range(read_matrix.shape[1])]

        # Fit quadratic x
        px = np.polyfit(range(0, len(medians_x)), medians_x, 2)
        LOGGER.debug(f"[{self.filename}] : x polyfit 2nd order: {px}")

        # Handle divide by zero
        if px[0] != 0:
            if not np.isnan(px[0]):
                # Remove quadratic in x
                cx = -px[1] / (2 * px[0])
                for row in range(0, image.shape[0]):
                    for col in range(0, image.shape[1]):
                        image[row, col] -= px[0] * (col - cx) ** 2
            else:
                LOGGER.debug(f"[{self.filename}] : Quadratic polyfit returns nan, skipping quadratic removal")
        else:
            LOGGER.debug(f"[{self.filename}] : Quadratic polyfit returns zero, skipping quadratic removal")

        return image

    @staticmethod
    def calc_diff(array: npt.NDArray) -> npt.NDArray:
        """
        Calculate the difference between the last and first rows of a 2-D array.

        Parameters
        ----------
        array : npt.NDArray
            A Numpy array.

        Returns
        -------
        npt.NDArray
            An array of the difference between the last and first rows of an array.
        """
        return array[-1] - array[0]

    def calc_gradient(self, array: npt.NDArray, shape: int) -> npt.NDArray:
        """
        Calculate the gradient of an array.

        Parameters
        ----------
        array : npt.NDArray
            Array for gradient to be calculated.
        shape : int
            Shape of the array.

        Returns
        -------
        npt.NDArray
            Gradient across the array.
        """
        return self.calc_diff(array) / shape

    def average_background(self, image: npt.NDArray, mask: npt.NDArray = None) -> npt.NDArray:
        """
        Zero the background by subtracting the non-masked mean from all pixels.

        Parameters
        ----------
        image : npt.NDArray
            Numpy array representing the image.
        mask : npt.NDArray
            Mask of the array, should have the same dimensions as image.

        Returns
        -------
        npt.NDArray
            Numpy array of image zero averaged.
        """
        if mask is None:
            mask = np.zeros_like(image)
        mean = np.mean(image[mask == 0])
        LOGGER.debug(f"[{self.filename}] : Zero averaging background : {mean} nm")
        return image - mean

    def gaussian_filter(self, image: npt.NDArray, **kwargs) -> npt.NDArray:
        """
        Apply Gaussian filter to an image.

        Parameters
        ----------
        image : npt.NDArray
            Numpy array representing the image.
        **kwargs
            Keyword arguments passed on to the skimage.filters.gaussian() function.

        Returns
        -------
        npt.NDArray
            Numpy array that represent the image after Gaussian filtering.
        """
        LOGGER.debug(
            f"[{self.filename}] : Applying Gaussian filter (mode : {self.gaussian_mode};"
            f" Gaussian blur (px) : {self.gaussian_size})."
        )
        return gaussian(
            image,
            sigma=(self.gaussian_size),
            mode=self.gaussian_mode,
            **kwargs,
        )

    def filter_image(self) -> None:  # numpydoc: ignore=GL07
        """
        Process a single image, filtering, finding grains and calculating their statistics.

        Returns
        -------
        None
            Does not return anything.

        Examples
        --------
        from topostats.io import LoadScan
        from topostats.filters import Filter
        from topostats.processing import process_scan

        filter = Filter(image=load_scan.image,
        ...             pixel_to_nm_scaling=load_scan.pixel_to_nm_scaling,
        ...             filename=load_scan.filename,
        ...             threshold_method='otsu')
        filter.filter_image()
        """
        self.images["initial_median_flatten"] = self.median_flatten(
            self.images["pixels"], mask=None, row_alignment_quantile=self.row_alignment_quantile
        )
        self.images["initial_tilt_removal"] = self.remove_tilt(self.images["initial_median_flatten"], mask=None)
        self.images["initial_quadratic_removal"] = self.remove_quadratic(self.images["initial_tilt_removal"], mask=None)
        self.images["initial_nonlinear_polynomial_removal"] = self.remove_nonlinear_polynomial(
            self.images["initial_quadratic_removal"], mask=None
        )

        # Remove scars
        run_scar_removal = self.remove_scars_config.pop("run")
        if run_scar_removal:
            LOGGER.debug(f"[{self.filename}] : Initial scar removal")
            self.images["initial_scar_removal"], _ = scars.remove_scars(
                self.images["initial_nonlinear_polynomial_removal"],
                filename=self.filename,
                **self.remove_scars_config,
            )
        else:
            LOGGER.debug(f"[{self.filename}] : Skipping scar removal as requested from config")
            self.images["initial_scar_removal"] = self.images["initial_nonlinear_polynomial_removal"]

        # Zero the data before thresholding, helps with absolute thresholding
        self.images["initial_zero_average_background"] = self.average_background(
            self.images["initial_scar_removal"], mask=None
        )

        # Get the thresholds
        try:
            self.thresholds = get_thresholds(
                image=self.images["initial_zero_average_background"],
                threshold_method=self.threshold_method,
                otsu_threshold_multiplier=self.otsu_threshold_multiplier,
                threshold_std_dev=self.threshold_std_dev,
                absolute=self.threshold_absolute,
            )
        except TypeError as type_error:
            raise type_error
        self.images["mask"] = get_mask(
            image=self.images["initial_zero_average_background"],
            thresholds=self.thresholds,
            img_name=self.filename,
        )
        self.images["masked_median_flatten"] = self.median_flatten(
            self.images["initial_tilt_removal"],
            self.images["mask"],
            row_alignment_quantile=self.row_alignment_quantile,
        )
        self.images["masked_tilt_removal"] = self.remove_tilt(self.images["masked_median_flatten"], self.images["mask"])
        self.images["masked_quadratic_removal"] = self.remove_quadratic(
            self.images["masked_tilt_removal"], self.images["mask"]
        )
        self.images["masked_nonlinear_polynomial_removal"] = self.remove_nonlinear_polynomial(
            self.images["masked_quadratic_removal"], self.images["mask"]
        )
        # Remove scars
        if run_scar_removal:
            LOGGER.debug(f"[{self.filename}] : Secondary scar removal")
            self.images["secondary_scar_removal"], scar_mask = scars.remove_scars(
                self.images["masked_nonlinear_polynomial_removal"],
                filename=self.filename,
                **self.remove_scars_config,
            )
            self.images["scar_mask"] = scar_mask
        else:
            LOGGER.debug(f"[{self.filename}] : Skipping scar removal as requested from config")
            self.images["secondary_scar_removal"] = self.images["masked_nonlinear_polynomial_removal"]
        self.images["final_zero_average_background"] = self.average_background(
            self.images["secondary_scar_removal"], self.images["mask"]
        )
        self.images["gaussian_filtered"] = self.gaussian_filter(self.images["final_zero_average_background"])



"""Function for calculating statistics about a whole image, for example number of grains or surface roughness."""

import logging

import numpy as np
import pandas as pd

from topostats.logs.logs import LOGGER_NAME

LOGGER = logging.getLogger(LOGGER_NAME)


def image_statistics(
    image: np.ndarray,
    filename: str,
    pixel_to_nm_scaling: float,
    results_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Calculate statistics pertaining to the whole image.

    Calculates the size of the image in pixels and metres, the root-mean-squared roughness and the grains per metre
    squared.

    Parameters
    ----------
    image : np.ndarray
        Numpy 2D image array of the image to calculate stats for.
    filename : str
        The name of the file being processed.
    pixel_to_nm_scaling : float
        Float of the scaling factor between pixels and nanometres.
    results_df : pd.DataFrame
        Pandas DataFrame of statistics pertaining to individual grains including from grainstats and
        dna tracing.

    Returns
    -------
    dict
        Dictionary of image statistics.
    """
    image_stats = {
        "image": filename,
        "image_size_x_m": None,
        "image_size_y_m": None,
        "image_area_m2": None,
        "image_size_x_px": image.shape[1],
        "image_size_y_px": image.shape[0],
        "image_area_px2": None,
        "grains_number_above": None,
        "grains_per_m2_above": None,
        "grains_number_below": None,
        "grains_per_m2_below": None,
        "rms_roughness": None,
    }

    # Calculate dimensions of the image
    image_stats["image_size_x_m"] = image.shape[1] * pixel_to_nm_scaling * 1e-9
    image_stats["image_size_y_m"] = image.shape[0] * pixel_to_nm_scaling * 1e-9
    image_stats["image_area_m2"] = image_stats["image_size_x_m"] * image_stats["image_size_y_m"]
    image_stats["image_area_px2"] = image_stats["image_size_x_px"] * image_stats["image_size_y_px"]

    # Calculate the RMS roughness of the sample on the flattened image.
    image_stats["rms_roughness"] = roughness_rms(image=image) * 1e-9

    # Calculate image stats relating to grain statistics. Note that the existence of any of these stats
    # is not guaranteed
    try:
        image_stats["grains_number_below"] = results_df["threshold"].value_counts().get("below", 0)
        image_stats["grains_per_m2_below"] = image_stats["grains_number_below"] / image_stats["image_area_m2"]
    except KeyError:
        pass
    try:
        image_stats["grains_number_above"] = results_df["threshold"].value_counts().get("above", 0)
        image_stats["grains_per_m2_above"] = image_stats["grains_number_above"] / image_stats["image_area_m2"]
    except KeyError:
        pass

    image_stats_df = pd.DataFrame([image_stats])
    image_stats_df.set_index("image", inplace=True)

    return image_stats_df


def roughness_rms(image: np.ndarray) -> float:
    """
    Calculate the root-mean-square roughness of a heightmap image.

    Parameters
    ----------
    image : np.ndarray
        2-D numpy array of heightmap data to calculate roughness.

    Returns
    -------
    float
        The RMS roughness of the input array.
    """
    return np.sqrt(np.mean(np.square(image)))
